# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-08-07

import logging
from contextlib import contextmanager

from sqlalchemy.orm.exc import ObjectDeletedError

logger = logging.getLogger(__name__)


class RekordboxAgentRegistry:
    """Rekordbox Agent Registry handler.

    The Rekordbox Agent Registry handler is responsible for tracking changes to the
    Rekordbox database. It is used to keep track of changes to the database,
    provide a history of changes and to update the global and individual
    USN (unique sequence number) values of the database entries.

    This object should *not* be instantiated directly. It is used by the
    :class:`RekordboxDatabase` class to track changes to the database.

    Parameters
    ----------
    db : RekordboxDatabase
        The Rekordbox database instance.
    """

    __update_sequence__ = list()
    __update_history__ = list()
    __enabled__ = True

    def __init__(self, db):
        self.db = db

    @classmethod
    def on_update(cls, instance, key, value):
        """Called when an instance of a database model is updated.

        Parameters
        ----------
        instance : tables.Base
            The table entry instance.
        key : str
            The name of the updated column.
        value : Any
            The new value of the updated column.
        """
        if cls.__enabled__:
            logger.debug("On update: %s, %s, %s", instance, key, value)
            cls.__update_sequence__.append((instance, "update", key, value))

    @classmethod
    def on_create(cls, instance):
        """Called when an instance of a database model is created.

        Parameters
        ----------
        instance : tables.Base
            The table entry instance.
        """
        if cls.__enabled__:
            logger.debug("On create: %s", instance)
            cls.__update_sequence__.append((instance, "create", "", ""))

    @classmethod
    def on_delete(cls, instance):
        """Called when an instance of a database model is deleted.

        Parameters
        ----------
        instance : tables.Base
            The table entry instance.
        """
        if cls.__enabled__:
            try:
                s = str(instance)
                logger.debug("On delete: %s", s)
            except ObjectDeletedError:
                instance = []

            cls.__update_sequence__.append((instance, "delete", "", ""))

    @classmethod
    def on_move(cls, instances):
        """Called when instanced of a database model are moved.

        Parameters
        ----------
        instances : list[tables.Base]
            The table entry instance.
        """
        if cls.__enabled__:
            logger.debug("On move: %s", instances)
            cls.__update_sequence__.append((instances, "move", "", ""))

    @classmethod
    def clear_buffer(cls):
        """Clears the update buffer and update history."""
        cls.__update_history__.extend(cls.__update_sequence__)
        cls.__update_sequence__.clear()

    @classmethod
    def enable_tracking(cls):
        """Enables the tracking of database changes."""
        cls.__enabled__ = True

    @classmethod
    def disable_tracking(cls):
        """Disables the tracking of database changes."""
        cls.__enabled__ = False

    @classmethod
    @contextmanager
    def disabled(cls):
        """Context manager to temporarily disable the tracking of database changes.

        Examples
        --------
        >>> registry = RekordboxAgentRegistry(db)  # noqa
        >>> registry.__enabled__
        True

        >>> with registry.disabled():
        ...     print(registry.__enabled__)
        False
        """
        enabled = cls.__enabled__
        cls.disable_tracking()
        yield cls
        if enabled:
            cls.enable_tracking()

    def get_registries(self):
        """Returns all agent registries.

        Returns
        -------
        registries : list[tables.AgentRegistry]
        """
        return self.db.get_agent_registry()

    def get_registry(self, key):
        """Returns the agent registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).

        Returns
        -------
        registry : tables.AgentRegistry
        """
        return self.db.get_agent_registry(registry_id=key)

    def get_string(self, key):
        """Returns the string value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).

        Returns
        -------
        value : str
        """
        return self.db.get_agent_registry(registry_id=key).str_1

    def get_text(self, key):
        """Returns the text value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).

        Returns
        -------
        value : str
        """
        return self.db.get_agent_registry(registry_id=key).text_1

    def get_int(self, key):
        """Returns the integer value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).

        Returns
        -------
        value : int
        """
        return self.db.get_agent_registry(registry_id=key).int_1

    def get_date(self, key):
        """Returns the date value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).

        Returns
        -------
        value : datetime.datetime
        """
        return self.db.get_agent_registry(registry_id=key).date_1

    def set_string(self, key, value):
        """Sets the string value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).
        value : str
            The new value.
        """
        self.db.get_agent_registry(registry_id=key).str_1 = value

    def set_text(self, key, value):
        """Sets the text value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).
        value : str
            The new value.
        """
        self.db.get_agent_registry(registry_id=key).text_1 = value

    def set_int(self, key, value):
        """Sets the integer value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).
        value : int
            The new value.
        """
        self.db.get_agent_registry(registry_id=key).int_1 = value

    def set_date(self, key, value):
        """Sets the date value of the registry with the given key.

        Parameters
        ----------
        key : str
            The registry identifier (`registry_id`).
        value : datetime.datetime
            The new value.
        """
        self.db.get_agent_registry(registry_id=key).date_1 = value

    def get_local_update_count(self):
        """Returns the current global local USN (unique sequence number)."""
        reg = self.db.get_agent_registry(registry_id="localUpdateCount")
        return reg.int_1

    def set_local_update_count(self, value):
        """Sets the global local USN (unique sequence number).

        Parameters
        ----------
        value : int
            The new USN value.
        """
        reg = self.db.get_agent_registry(registry_id="localUpdateCount")
        reg.int_1 = value

    def increment_local_update_count(self, num=1):
        """Increments the global local USN (unique sequence number) by the given number.

        Parameters
        ----------
        num : int, optional
            The number to increment the USN by. The default is 1.

        Returns
        -------
        usn: int
            The new global local USN.
        """
        if not isinstance(num, int) or num < 1:
            raise ValueError("The USN can only be increment by a positive integer!")
        reg = self.db.get_agent_registry(registry_id="localUpdateCount")
        reg.int_1 = reg.int_1 + num
        return reg.int_1

    def autoincrement_local_update_count(self, set_row_usn=True):
        """Auto-increments the global local USN (unique sequence number).

        The number of changes in the update buffer is used to determine the
        number to increment the USN by. After the update the buffer is cleared.

        Parameters
        ----------
        set_row_usn : bool, optional
            If True the local USN of each database entry is updated with the
            corresponding value in the order the changes were made.

        Returns
        -------
        usn: int
            The new global local USN.
        """
        reg = self.db.get_agent_registry(registry_id="localUpdateCount")
        usn = reg.int_1
        self.disable_tracking()
        self.db.flush()
        with self.db.session.no_autoflush:
            for instances, op, _, _ in self.__update_sequence__.copy():
                usn += 1
                if set_row_usn:
                    # All instances in a list get the same USN
                    if not isinstance(instances, list):
                        instances = [instances]
                    for instance in instances:
                        if hasattr(instance, "rb_local_usn"):
                            instance.rb_local_usn = usn
            reg.int_1 = usn

        self.clear_buffer()
        self.db.flush()
        self.enable_tracking()
        return usn



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-08-15

import os
import re
import shutil
import sys
import urllib.request
from pathlib import Path

from pyrekordbox.config import get_cache_file, write_db6_key_cache

KEY_SOURCES = [
    {
        "url": r"https://raw.githubusercontent.com/mganss/CueGen/19878e6eb3f586dee0eb3eb4f2ce3ef18309de9d/CueGen/Generator.cs",  # noqa: E501
        "regex": re.compile(
            r'((.|\n)*)Config\.UseSqlCipher.*\?.*"(?P<dp>.*)".*:.*null',
            flags=re.IGNORECASE | re.MULTILINE,
        ),
    },
    {
        "url": r"https://raw.githubusercontent.com/dvcrn/go-rekordbox/8be6191ba198ed7abd4ad6406d177ed7b4f749b5/cmd/getencryptionkey/main.go",  # noqa: E501
        "regex": re.compile(
            r'((.|\n)*)fmt\.Print\("(?P<dp>.*)"\)', flags=re.IGNORECASE | re.MULTILINE
        ),
    },
]


class WorkingDir:
    def __init__(self, path):
        self._prev = Path.cwd()
        self.path = Path(path)

    def __enter__(self):
        if self.path != self._prev:
            self.path.mkdir(parents=True, exist_ok=True)
            os.chdir(self.path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.path != self._prev:
            os.chdir(self._prev)


def clone_repo(https_url: str) -> Path:
    path = Path.cwd() / https_url.split("/")[-1]
    if not path.exists():
        os.system(f"git clone {https_url}")
        assert path.exists()
    else:
        print(f"Repository {https_url} already cloned")
    return path


def clone_pysqlcipher3() -> Path:
    return clone_repo(r"https://github.com/coleifer/sqlcipher3")


def clone_sqlcipher_amalgamation() -> Path:
    return clone_repo(r"https://github.com/geekbrother/sqlcipher-amalgamation")


def patch_pysqlcipher_setup(pysqlcipher_dir, cryptolib="libcrypto.lib"):
    path = Path(pysqlcipher_dir, "setup.py")

    with open(path, "r") as fh:
        text = fh.read()

    if cryptolib:
        lib_old = "os.environ.get('OPENSSL_LIBNAME') or 'libeay32.lib'"
        lib_new = f"os.environ.get('OPENSSL_LIBNAME') or '{cryptolib}'"
        text = text.replace(lib_old, lib_new)

    with open(path, "w") as fh:
        fh.write(text)


def prepare_pysqlcipher(pysqlcipher_dir: Path, amalgamation_src: Path):
    # Copy amalgamation files to pysqlcipher directory
    root = pysqlcipher_dir
    root.mkdir(parents=True, exist_ok=True)
    shutil.copy2(amalgamation_src / "sqlite3.c", root / "sqlite3.c")
    shutil.copy2(amalgamation_src / "sqlite3.h", root / "sqlite3.h")


def install_pysqlcipher(
    tmpdir="pysqlcipher3",
    crypto_lib="libcrypto.lib",
    pyexecutable="",
    build=True,
    install=True,
    cleanup=True,
):
    tmpdir = Path(tmpdir)
    # Download pysqlcipher3 and prepare amalgamation build
    with WorkingDir(tmpdir):
        pysqlcipher_dir = clone_pysqlcipher3()
        amalgamation_dir = clone_sqlcipher_amalgamation()
        amalgamation_src = amalgamation_dir / "src"

        prepare_pysqlcipher(pysqlcipher_dir, amalgamation_src)
        if sys.platform == "win32" and os.getenv("OPENSSL_LIBNAME") is None:
            print("No OPENSSL_LIBNAME environment variable found, updating `setup.py`!")
            patch_pysqlcipher_setup(pysqlcipher_dir, crypto_lib)

    # Build amalgamation and install pysqlcipher
    if not pyexecutable:
        pyexecutable = sys.executable

    with WorkingDir(pysqlcipher_dir):
        if build:
            # Build amalgamation
            print()
            os.system(f"{pyexecutable} setup.py build_static build")
        if install:
            # Install pysqlcipher package
            print()
            os.system(f"{pyexecutable} setup.py install")

    # Remove temporary files
    if cleanup:
        try:
            print()
            print("Cleaning up")
            tmpdir.unlink(missing_ok=True)
        except PermissionError as e:
            print()
            print(e)
            print(f"Could not remove temporary directory '{tmpdir}'!")


def download_db6_key():
    dp = ""
    for source in KEY_SOURCES:
        url = source["url"]
        regex = source["regex"]
        print(f"Looking for key: {url}")

        res = urllib.request.urlopen(url)
        data = res.read().decode("utf-8")
        match = regex.match(data)
        if match:
            dp = match.group("dp")
            break
    if dp:
        cache_file = get_cache_file()
        print(f"Found key, updating cache file {cache_file}")
        write_db6_key_cache(dp)
    else:
        print("No key found in the online sources.")


def main():
    from argparse import ArgumentParser

    parser = ArgumentParser("pyrekordbox")
    subparsers = parser.add_subparsers(dest="command")

    # Download Rekordbx 6 database key command
    subparsers.add_parser(
        "download-key",
        help="Download the Rekordbox 6 database key from the internet "
        "and write it to the cache file.",
    )

    # Install pysqlcipher3 command (Windows only)
    install_parser = subparsers.add_parser(
        "install-sqlcipher",
        help="Build sqlcipher against amalgamation and install pysqlcipher3",
    )
    install_parser.add_argument(
        "-t",
        "--tmpdir",
        type=str,
        default=".tmp",
        help="Path for storing temporary data (default: '.tmp')",
    )
    install_parser.add_argument(
        "-l",
        "--cryptolib",
        type=str,
        default="libcrypto.lib",
        help="The name of the OpenSSl crypto libary (default: 'libcrypto.lib')",
    )
    install_parser.add_argument(
        "-b",
        "--buildonly",
        action="store_true",
        help="Don't install sqlcipher3, only build the amalgamation",
    )

    # Parse args and handle command
    args = parser.parse_args(sys.argv[1:])
    if args.command == "download-key":
        download_db6_key()
    elif args.command == "install-sqlcipher":
        install_pysqlcipher(args.tmpdir, args.cryptolib, install=not args.buildonly)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-10-02

import json
import sys
from pathlib import Path

import pytest

from pyrekordbox.config import get_config, update_config

RB_SETTING = """<?xml version="1.0" encoding="UTF-8"?>
<PROPERTIES><VALUE name="masterDbDirectory" val="{db_dir}"/></PROPERTIES>
"""

RB_OPTIONS = {"options": [["db-path", ""]]}


def mock_rekordbox_settings_file(pioneer_app_dir, rb_app_dir):
    db_dir = pioneer_app_dir / "rekordbox"
    text = RB_SETTING.format(db_dir=db_dir)
    file = rb_app_dir / "rekordbox3.settings"
    file.write_text(text)


def mock_rekordbox_options_file(agent_storage_dir, db_path):
    file = agent_storage_dir / "options.json"
    options = RB_OPTIONS.copy()
    options["options"][0][1] = str(db_path)
    with open(file, "w") as fp:
        json.dump(options, fp)


@pytest.fixture(scope="function")
def pioneer_app_dir(tmp_path_factory):
    root = tmp_path_factory.mktemp("app")
    rb_dir = root / "rekordbox"
    rb6_dir = root / "rekordbox6"
    agent_storage_dir = root / "rekordboxAgent" / "storage"
    rb_dir.mkdir(parents=True)
    rb6_dir.mkdir(parents=True)
    agent_storage_dir.mkdir(parents=True)

    # Mock `rekordbox3.settings` files
    mock_rekordbox_settings_file(root, rb_dir)
    mock_rekordbox_settings_file(root, rb6_dir)

    # Mock database files
    rb5_db = rb_dir / "datafile.edb"
    rb6_db = rb_dir / "master.db"
    rb5_db.touch()
    rb6_db.touch()

    # Mock `options.json` file (RBv6)
    mock_rekordbox_options_file(agent_storage_dir, rb6_db)

    return root


@pytest.fixture(scope="function")
def pioneer_install_dir(tmp_path_factory):
    root = tmp_path_factory.mktemp("install")
    if sys.platform == "win32":
        root_56 = root / "Pioneer"
        root_7 = root / "rekordbox"
    else:
        root_56 = root
        root_7 = root
    rb5_dir = root_56 / "rekordbox 5.1.1"
    rb6_dir = root_56 / "rekordbox 6.1.1"
    rb7_dir = root_7 / "rekordbox 7.1.1"
    rb5_dir_alt = root_56 / "rekordbox 5.1.2"
    rb6_dir_alt = root_56 / "rekordbox 6.1.2"
    rb7_dir_alt = root_7 / "rekordbox 7.1.2"

    rb5_dir.mkdir(parents=True)
    rb6_dir.mkdir(parents=True)
    rb7_dir.mkdir(parents=True)
    rb5_dir_alt.mkdir(parents=True)
    rb6_dir_alt.mkdir(parents=True)
    rb7_dir_alt.mkdir(parents=True)
    return root


def test_pioneer_config(pioneer_install_dir, pioneer_app_dir):
    update_config(pioneer_install_dir, pioneer_app_dir)
    install_dir = get_config("pioneer", "install_dir")
    app_dir = get_config("pioneer", "app_dir")

    assert isinstance(install_dir, Path)
    assert isinstance(app_dir, Path)
    assert install_dir == pioneer_install_dir
    assert app_dir == pioneer_app_dir


def test_rb5_config(pioneer_install_dir, pioneer_app_dir):
    update_config(pioneer_install_dir, pioneer_app_dir)
    expected_version = "5.1.2"
    if sys.platform == "win32":
        pioneer_install_dir = pioneer_install_dir / "Pioneer"

    app_dir = get_config("rekordbox5", "app_dir")
    install_dir = get_config("rekordbox5", "install_dir")
    db_dir = get_config("rekordbox5", "db_dir")
    db_path = get_config("rekordbox5", "db_path")
    version = get_config("rekordbox5", "version")
    assert isinstance(install_dir, Path)
    assert isinstance(app_dir, Path)
    assert isinstance(db_dir, Path)
    assert isinstance(db_path, Path)
    assert install_dir == (pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox")
    assert db_dir == (pioneer_app_dir / "rekordbox")
    assert db_path == (pioneer_app_dir / "rekordbox" / "datafile.edb")
    assert version == expected_version


def test_rb6_config(pioneer_install_dir, pioneer_app_dir):
    update_config(pioneer_install_dir, pioneer_app_dir)
    expected_version = "6.1.2"
    if sys.platform == "win32":
        pioneer_install_dir = pioneer_install_dir / "Pioneer"

    app_dir = get_config("rekordbox6", "app_dir")
    install_dir = get_config("rekordbox6", "install_dir")
    db_dir = get_config("rekordbox6", "db_dir")
    db_path = get_config("rekordbox6", "db_path")
    version = get_config("rekordbox6", "version")

    assert isinstance(install_dir, Path)
    assert isinstance(app_dir, Path)
    assert isinstance(db_dir, Path)
    assert isinstance(db_path, Path)
    assert install_dir == (pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox6")
    assert db_dir == (pioneer_app_dir / "rekordbox")
    assert db_path == (pioneer_app_dir / "rekordbox" / "master.db")
    assert version == expected_version


def test_rb7_config(pioneer_install_dir, pioneer_app_dir):
    update_config(pioneer_install_dir, pioneer_app_dir)
    expected_version = "7.1.2"
    if sys.platform == "win32":
        pioneer_install_dir = pioneer_install_dir / "rekordbox"

    app_dir = get_config("rekordbox7", "app_dir")
    install_dir = get_config("rekordbox7", "install_dir")
    db_dir = get_config("rekordbox7", "db_dir")
    db_path = get_config("rekordbox7", "db_path")
    version = get_config("rekordbox7", "version")

    assert isinstance(install_dir, Path)
    assert isinstance(app_dir, Path)
    assert isinstance(db_dir, Path)
    assert isinstance(db_path, Path)
    assert install_dir == (pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox6")
    assert db_dir == (pioneer_app_dir / "rekordbox")
    assert db_path == (pioneer_app_dir / "rekordbox" / "master.db")
    assert version == expected_version


def test_rb5_config_alt(pioneer_install_dir, pioneer_app_dir):
    # test default: latest version
    update_config(pioneer_install_dir, pioneer_app_dir)
    expected_version = "5.1.2"
    if sys.platform == "win32":
        full_pioneer_install_dir = pioneer_install_dir / "Pioneer"
    else:
        full_pioneer_install_dir = pioneer_install_dir

    app_dir = get_config("rekordbox5", "app_dir")
    install_dir = get_config("rekordbox5", "install_dir")
    version = get_config("rekordbox5", "version")
    assert install_dir == (full_pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox")
    assert version == expected_version

    # test alternative version
    update_config(pioneer_install_dir, pioneer_app_dir, rb5_install_dirname="rekordbox 5.1.1")
    expected_version = "5.1.1"

    app_dir = get_config("rekordbox5", "app_dir")
    install_dir = get_config("rekordbox5", "install_dir")
    version = get_config("rekordbox5", "version")
    assert install_dir == (full_pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox")
    assert version == expected_version


def test_rb6_config_alt(pioneer_install_dir, pioneer_app_dir):
    # test default: latest version
    update_config(pioneer_install_dir, pioneer_app_dir)
    expected_version = "6.1.2"
    if sys.platform == "win32":
        full_pioneer_install_dir = pioneer_install_dir / "Pioneer"
    else:
        full_pioneer_install_dir = pioneer_install_dir

    app_dir = get_config("rekordbox6", "app_dir")
    install_dir = get_config("rekordbox6", "install_dir")
    version = get_config("rekordbox6", "version")
    assert install_dir == (full_pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox6")
    assert version == expected_version

    # test alternative version
    update_config(pioneer_install_dir, pioneer_app_dir, rb6_install_dirname="rekordbox 6.1.1")
    expected_version = "6.1.1"

    app_dir = get_config("rekordbox6", "app_dir")
    install_dir = get_config("rekordbox6", "install_dir")
    version = get_config("rekordbox6", "version")
    assert install_dir == (full_pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox6")
    assert version == expected_version


def test_rb7_config_alt(pioneer_install_dir, pioneer_app_dir):
    # test default: latest version
    update_config(pioneer_install_dir, pioneer_app_dir)
    expected_version = "7.1.2"
    if sys.platform == "win32":
        full_pioneer_install_dir = pioneer_install_dir / "rekordbox"
    else:
        full_pioneer_install_dir = pioneer_install_dir

    app_dir = get_config("rekordbox7", "app_dir")
    install_dir = get_config("rekordbox7", "install_dir")
    version = get_config("rekordbox7", "version")
    assert install_dir == (full_pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox6")
    assert version == expected_version

    # test alternative version
    update_config(pioneer_install_dir, pioneer_app_dir, rb7_install_dirname="rekordbox 7.1.1")
    expected_version = "7.1.1"

    app_dir = get_config("rekordbox7", "app_dir")
    install_dir = get_config("rekordbox7", "install_dir")
    version = get_config("rekordbox7", "version")
    assert install_dir == (full_pioneer_install_dir / f"rekordbox {expected_version}")
    assert app_dir == (pioneer_app_dir / "rekordbox6")
    assert version == expected_version



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2022-04-10

"""Configuration handling for pyrekordbox.

Contains all the path and settings handling of the Rekordbox installation(s) on
the users machine.
"""

import base64
import json
import logging
import os
import re
import sys
import textwrap
import time
import xml.etree.cElementTree as xml
from pathlib import Path
from typing import Union

import blowfish
import frida
import packaging.version

from .utils import get_rekordbox_pid

logger = logging.getLogger(__name__)

# Cache file for pyrekordbox data
_cache_file_version = 2
_cache_file_name = "rb.cache"

# Define empty pyrekordbox configuration
__config__ = {
    "pioneer": {
        "app_dir": Path(),
        "install_dir": Path(),
    },
    "rekordbox5": {},
    "rekordbox6": {},
    "rekordbox7": {},
}


class InvalidApplicationDirname(Exception):
    pass


def get_appdata_dir() -> Path:
    """Returns the path of the application data directory.

    On Windows, the application data is stored in `/Users/user/AppData/Roaming`.
    On macOS the application data is stored in `~/Libary/Application Support`.
    """
    if sys.platform == "win32":
        # Windows: located in /Users/user/AppData/Roaming/
        app_data = Path(os.environ["AppData"])
    elif sys.platform == "darwin":
        # MacOS: located in ~/Library/Application Support/
        app_data = Path("~").expanduser() / "Library" / "Application Support"
    else:
        # Linux: not supported
        logger.warning(f"OS {sys.platform} not supported!")
        return Path("~").expanduser() / ".local" / "share"
    return app_data


def get_cache_file() -> Path:
    return get_appdata_dir() / "pyrekordbox" / _cache_file_name


def get_pioneer_install_dir(path: Union[str, Path] = None) -> Path:  # pragma: no cover
    """Returns the path of the Pioneer program installation directory.

    On Windows, the Pioneer program data is stored in `/ProgramFiles/Pioneer`.
    For rekordbox version 7 this has changed to `/ProgramFiles/rekordbox`.
    On macOS the program data is somewhere in `/Applications/`.

    Parameters
    ----------
    path : str or Path, optional
        If a path is given it will only be checked for validity. Otherwise,
        the default Pioneer directory will be constructed and checked.

    Returns
    -------
    pioneer_path : Path
        The path to the Pioneer program installation data.
    """
    if path is None:
        if sys.platform == "win32":
            # Windows: located in /ProgramFiles/Pioneer
            program_files = os.environ["ProgramFiles"].replace("(x86)", "").strip()
            path = Path(program_files)
        elif sys.platform == "darwin":
            # MacOS: located in /Applications/
            path = Path("/Applications")
        else:
            # Linux: not supported
            logger.warning(f"OS {sys.platform} not supported!")
            return Path()
    else:
        path = Path(path)

    path = path.absolute()
    if not path.exists():
        raise FileNotFoundError(f"The Pioneer install-dir {path} does not exist!")

    return path


def get_pioneer_app_dir(path: Union[str, Path] = None) -> Path:  # pragma: no cover
    """Returns the path of the Pioneer application data directory.

    On Windows, the Pioneer application data is stored in `/Users/user/AppData/Roaming`
    On macOS the application data is somewhere in `~/Libary/Application Support`.

    Parameters
    ----------
    path : str or Path, optional
        If a path is given it will only be checked for validity. Otherwise,
        the default Pioneer directory will be constructed and checked.

    Returns
    -------
    pioneer_path : Path
        The path to the Pioneer application data.
    """
    if path is None:
        if sys.platform == "win32":
            # Windows: located in /Users/user/AppData/Roaming/
            app_data = Path(os.environ["AppData"])
        elif sys.platform == "darwin":
            # MacOS: located in ~/Library/Application Support/
            app_data = Path("~").expanduser() / "Library" / "Application Support"
        else:
            # Linux: not supported
            logger.warning(f"OS {sys.platform} not supported!")
            return Path()
        # Pioneer app data
        path = app_data / "Pioneer"
    else:
        path = Path(path)

    path = path.absolute()
    if not path.exists():
        raise FileNotFoundError(f"The Pioneer app-dir {path} does not exist!")

    return path


def _convert_type(s):
    # Try to parse as int, float, list of int, list of float
    types_ = int, float
    for type_ in types_:
        try:
            return type_(s)
        except ValueError:
            pass
        try:
            return [type_(x) for x in s.split(",")]
        except ValueError:
            pass

    return s


def read_rekordbox_settings(rekordbox_app_dir: Union[str, Path]) -> dict:
    """Finds and parses the 'rekordbox3.settings' file in the Rekordbox 5 or 6 app-dir.

    The settings file usually is called 'rekordbox3.settings' and is
    located in the application data directory of the corresponding Rekordbox
    (major) version.

    Parameters
    ----------
    rekordbox_app_dir : str or Path
        The path of the application-data directory of Rekordbox 5 or 6.

    Returns
    -------
    settings : dict
        The parsed Rekordbox settings data.
    """
    # Get path of the settings file
    rekordbox_app_dir = Path(rekordbox_app_dir)
    path = rekordbox_app_dir / "rekordbox3.settings"

    # Parse the settings file
    settings = dict()
    tree = xml.parse(path)
    for element in tree.findall("VALUE"):
        name = element.attrib["name"]
        try:
            val = _convert_type(element.attrib["val"])
        except KeyError:
            device_setup = element.find("DEVICESETUP")
            val = {k: _convert_type(v) for k, v in device_setup.attrib.items()}
        settings[name] = val
    return settings


def read_rekordbox6_options(pioneer_app_dir: Union[str, Path]) -> dict:
    """Finds and parses the Rekordbox 6 `options.json` file with additional settings.

    The options file contains additional settings used by Rekordbox 6, for example the
    path of the new `master.db` database. It also contains some data nedded to open
    the database, which is encrypted using SQLCipher.

    Parameters
    ----------
    pioneer_app_dir : str or Path, optional
        The path of the Pioneer application data.

    Returns
    -------
    options : dict
        The parsed rekordbox 6 options data.
    """
    # Get path of the options file
    pioneer_app_dir = Path(pioneer_app_dir)
    opt_path = pioneer_app_dir / "rekordboxAgent" / "storage" / "options.json"
    # Read and parse the options file
    with open(opt_path, "r") as fh:
        data = json.load(fh)
    options = dict()
    for key, value in data["options"]:
        options[key] = value
    return options


def read_rekordbox6_asar(rb6_install_dir: Union[str, Path]) -> str:
    """Finds and parses the Rekordbox 6 `app.asar` archive file.

    An ASAR file is an archive used to package source code for an application
    using Electron. Rekordbox 6 stores some useful information for opening the
    new `master.db` database in this file.

    Parameters
    ----------
    rb6_install_dir : str or Path
        The path of the Rekordbox 6 installation directory.

    Returns
    -------
    asar_data : str
        The data of the Rekordbox 6 `app.asar` archive file as ANSI encoded string.
    """
    rb6_install_dir = Path(rb6_install_dir)
    # Find location of app asar file
    if sys.platform == "win32":
        location = rb6_install_dir / "rekordboxAgent-win32-x64" / "resources"
        encoding = "ANSI"
    elif sys.platform == "darwin":
        if not str(rb6_install_dir).endswith(".app"):
            rb6_install_dir = rb6_install_dir / "rekordbox.app"
        location = (
            rb6_install_dir / "Contents" / "MacOS" / "rekordboxAgent.app" / "Contents" / "Resources"
        )
        encoding = "cp437"
    else:
        logger.warning(f"OS {sys.platform} not supported!")
        return ""

    # Read asar file
    path = (location / "app.asar").absolute()
    with open(path, "rb") as fh:
        data = fh.read().decode(encoding, errors="replace")
    return data


def _extract_version(name, major_version):
    name = name.replace(".app", "")  # Needed for MacOS
    ver_str = name.replace("rekordbox", "").strip()
    if not ver_str:
        ver_str = str(major_version)
    return ver_str


def _get_rb_config(
    pioneer_install_dir: Path,
    pioneer_app_dir: Path,
    major_version: int,
    application_dirname: str = "",
) -> dict:
    """Get the program configuration for a given Rekordbox major version.

    Parameters
    ----------
    pioneer_install_dir : Path
        The path of the Pioneer installation directory.
    pioneer_app_dir : Path
        The path of the Pioneer application data directory.
    major_version : int
        The major version of Rekordbox.
    application_dirname : str, optional
        The name of the Rekordbox application directory. If not given, the latest
        Rekordbox installation directory for major version `major_version` is used.

    Returns
    -------
    config : dict
        The program configuration.
    """
    if sys.platform == "win32":
        if major_version >= 7:
            pioneer_install_dir = pioneer_install_dir / "rekordbox"
        else:
            pioneer_install_dir = pioneer_install_dir / "Pioneer"

    if application_dirname:
        # Applitcation dirname is given, only extract version from it
        # `major_version` is compared to the version string
        rb_prog_dir = pioneer_install_dir / application_dirname
        if not rb_prog_dir.exists():
            raise InvalidApplicationDirname(
                f"The supplied application dirname '{application_dirname}' does not "
                f"exist in '{pioneer_install_dir}'"
            )
        rb_version = _extract_version(application_dirname, major_version)
    else:
        # Get latest Rekordbox installation directory for major release `major_version`

        # Find all 'V.x.x' version strings in dir names
        installations = {}
        for p in pioneer_install_dir.iterdir():
            name = p.name
            if name.startswith("rekordbox"):
                ver_str = _extract_version(name, major_version)
                if ver_str.startswith(str(major_version)):
                    version = packaging.version.parse(ver_str)
                    installations[version] = name
        # Get latest 'V.x.x' version string and assure there is one
        versions = list(installations.keys())
        versions.sort()
        try:
            version = versions[-1]
        except IndexError:
            raise FileNotFoundError(
                f"No Rekordbox {major_version} folder found in installation "
                f"directory '{pioneer_install_dir}'"
            )
        # Name of the Rekordbox application directory in `pioneer_install_dir`
        rb_version = str(versions[-1])
        rb_prog_dir = pioneer_install_dir / installations[version]

    # Check installation directory
    if not rb_prog_dir.exists():
        raise FileNotFoundError(
            f"The Rekordbox installation directory '{rb_prog_dir}' doesn't exist"
        )
    logger.debug("Found Rekordbox %s install-dir: '%s'", major_version, rb_prog_dir)

    # Get Rekordbox application directory path for major release `major_version`
    name = "rekordbox6" if major_version >= 6 else "rekordbox"
    rb_app_dir = pioneer_app_dir / name
    if not rb_app_dir.exists():
        raise FileNotFoundError(f"The directory '{rb_app_dir}' doesn't exist!")
    logger.debug("Found Rekordbox %s app-dir: %s", major_version, rb_app_dir)

    # Get Rekordbox database locations for major release `major_version`
    settings = read_rekordbox_settings(rb_app_dir)
    db_dir = Path(settings["masterDbDirectory"])
    db_filename = "master.db" if major_version >= 6 else "datafile.edb"
    db_path = db_dir / db_filename
    if not db_path.exists():
        raise FileNotFoundError(f"The Rekordbox database '{db_path}' doesn't exist!")

    conf = {
        "version": rb_version,
        "install_dir": rb_prog_dir,
        "app_dir": rb_app_dir,
        "db_dir": db_dir,
        "db_path": db_path,
    }
    return conf


def _get_rb5_config(pioneer_prog_dir: Path, pioneer_app_dir: Path, dirname: str = "") -> dict:
    """Get the program configuration for Rekordbox v5.x.x."""
    major_version = 5
    conf = _get_rb_config(pioneer_prog_dir, pioneer_app_dir, major_version, dirname)
    return conf


def _extract_pw(pioneer_install_dir: Path) -> str:  # pragma: no cover
    """Extract the password for decrypting the Rekordbox 6 database key."""
    asar_data = read_rekordbox6_asar(pioneer_install_dir)
    match_result = re.search('pass: ".(.*?)"', asar_data)
    if match_result is None:
        raise RuntimeError("Could not read `app.asar` file.")
    match = match_result.group(0)
    pw = match.replace("pass: ", "").strip('"')
    return pw


class KeyExtractor:
    """Extracts the Rekordbox database key using code injection.

    This method works by injecting code into the Rekordbox process and intercepting the
    call to unlock the database. Works for any Rekordbox version.
    """

    # fmt: off
    SCRIPT = textwrap.dedent("""
    var sqlite3_key = Module.findExportByName(null, 'sqlite3_key');

    Interceptor.attach(sqlite3_key, {
        onEnter: function(args) {
            var size = args[2].toInt32();
            var key = args[1].readUtf8String(size);
            send('sqlite3_key: ' + key);
        }
    });
    """)
    # fmt: on

    def __init__(self, rekordbox_executable):
        self.executable = str(rekordbox_executable)
        self.key = ""

    def on_message(self, message, data):
        payload = message["payload"]
        if payload.startswith("sqlite3_key"):
            self.key = payload.split(": ")[1]

    def run(self):
        pid = get_rekordbox_pid()
        if pid:
            raise RuntimeError(
                "Rekordbox is running. Please close Rekordbox before running the `KeyExtractor`."
            )
        # Spawn Rekordbox process and attach to it
        pid = frida.spawn(self.executable)
        frida.resume(pid)
        session = frida.attach(pid)
        script = session.create_script(self.SCRIPT)
        script.on("message", self.on_message)
        script.load()
        # Wait for key to be extracted
        while not self.key:
            time.sleep(0.1)
        # Kill Rekordbox process
        frida.kill(pid)

        return self.key


def write_db6_key_cache(key: str) -> None:  # pragma: no cover
    r"""Writes the decrypted Rekordbox6 database key to the cache file.

    This method can also be used to manually cache the database key, provided
    the user has found the key somewhere else. The key can be, for example,
    found in some other projects that hard-coded it.

    The cache file is stored in the application data directory of pyrekordbox:
    Windows: `C:\Users\<user>\AppData\Roaming\pyrekordbox`
    macOS: `~/Library/Application Support/pyrekordbox`

    Parameters
    ----------
    key : str
        The decrypted database key. To make sure the key is valid, the first
        five characters are checked before writing the key to the cache file.
        The key should start with '402fd'.

    Examples
    --------
    >>> from pyrekordbox.config import write_db6_key_cache
    >>> from pyrekordbox import Rekordbox6Database
    >>> write_db6_key_cache("402fd...")
    >>> db = Rekordbox6Database()  # The db can now be opened without providing the key
    """
    # Check if the key looks like a valid key
    if not key.startswith("402fd"):
        raise ValueError("The provided database key doesn't look valid!")
    lines = list()
    lines.append(f"version: {_cache_file_version}")
    lines.append("dp: " + key)
    text = "\n".join(lines)

    cache_file = get_cache_file()
    if not cache_file.parent.exists():
        cache_file.parent.mkdir()

    with open(cache_file, "w") as fh:
        fh.write(text)
    # Set the config key to make sure the key is present after calling method
    if __config__["rekordbox6"]:
        __config__["rekordbox6"]["dp"] = key
    if __config__["rekordbox7"]:
        __config__["rekordbox7"]["dp"] = key


def _update_sqlite_key(opts, conf):
    cache_version = 0
    pw, dp = "", ""

    cache_file = get_cache_file()

    if cache_file.exists():  # pragma: no cover
        logger.debug("Found cache file %s", cache_file)
        # Read cache file
        with open(cache_file, "r") as fh:
            text = fh.read()
        lines = text.splitlines()
        if lines[0].startswith("version:"):
            cache_version = int(lines[0].split(":")[1].strip())
        else:
            cache_version = 1
        if cache_version == 1:
            # Cache file introduced in pyrekordbox 0.1.6 contains only the password
            pw = lines[0]
            logger.debug("Found pw in cache file")
        elif cache_version == 2:
            # Cache file introduced in pyrekordbox 0.1.7 contains version and db key
            dp = lines[1].split(":")[1].strip()
            logger.debug("Found dp in cache file")
        else:
            raise ValueError(f"Invalid cache version: {cache_version}")
    else:
        logger.debug("No cache file found")

    if cache_version < _cache_file_version:  # pragma: no cover
        # Update cache file
        if not pw:
            logger.debug("Extracting pw")
            try:
                pw = _extract_pw(conf["install_dir"])
                logger.debug("Extracted pw from 'app.asar'")
            except (FileNotFoundError, RuntimeError):
                logger.debug("Could not extract pw from 'app.asar'")
                pw = ""

        if not dp:
            if pw:
                cipher = blowfish.Cipher(pw.encode())
                dp = base64.standard_b64decode(opts["dp"])
                dp = b"".join(cipher.decrypt_ecb(dp)).decode()
                logger.debug("Unlocked dp from pw: %s", dp)
            else:
                if sys.platform == "win32":
                    executable = conf["install_dir"] / "rekordbox.exe"
                elif sys.platform == "darwin":
                    install_dir = conf["install_dir"]
                    if not str(install_dir).endswith(".app"):
                        install_dir = install_dir / "rekordbox.app"
                    executable = install_dir / "Contents" / "MacOS" / "rekordbox"
                else:
                    # Linux: not supported
                    logger.warning(f"OS {sys.platform} not supported!")
                    executable = Path("")

                if not executable.exists():
                    logger.warning(f"Could not find Rekordbox executable: {executable}")
                else:
                    extractor = KeyExtractor(executable)
                    try:
                        dp = extractor.run()
                        logger.debug("Extracted dp from Rekordbox process")
                    except Exception as e:
                        logger.info(f"`KeyExtractor` failed: {e}")

        if dp:
            logger.debug("Writing dp to cache file")
            write_db6_key_cache(dp)
        else:
            logging.warning(
                "Could not retrieve db-key or read it from cache file,"
                "use the CLI to download it from external sources: "
                "`python -m pyrekordbox download-key`"
            )

    # Add database key to config if found
    if dp:
        conf["dp"] = dp


def _get_rb6_config(pioneer_prog_dir: Path, pioneer_app_dir: Path, dirname: str = "") -> dict:
    """Get the program configuration for Rekordbox v6.x.x."""
    major_version = 6
    conf = _get_rb_config(pioneer_prog_dir, pioneer_app_dir, major_version, dirname)

    # Read Rekordbox 6 'options.json' and check db_path
    opts = read_rekordbox6_options(pioneer_app_dir)
    db_path = Path(opts["db-path"])
    db_dir = db_path.parent
    assert str(conf["db_dir"]) == str(db_dir)
    assert str(conf["db_path"]) == str(db_path)

    # Update SQLite key
    _update_sqlite_key(opts, conf)

    return conf


def _get_rb7_config(pioneer_prog_dir: Path, pioneer_app_dir: Path, dirname: str = "") -> dict:
    """Get the program configuration for Rekordbox v7.x.x."""
    major_version = 7
    conf = _get_rb_config(pioneer_prog_dir, pioneer_app_dir, major_version, dirname)

    # Read Rekordbox 6 'options.json' and check db_path
    opts = read_rekordbox6_options(pioneer_app_dir)
    db_path = Path(opts["db-path"])
    db_dir = db_path.parent
    assert str(conf["db_dir"]) == str(db_dir)
    assert str(conf["db_path"]) == str(db_path)

    # Update SQLite key
    _update_sqlite_key(opts, conf)

    return conf


# noinspection PyPackageRequirements,PyUnresolvedReferences
def _read_config_file(path: str) -> dict:
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"No such file or directory: '{path}'")

    ext = path.suffix.lower()
    if ext == ".cfg":
        from configparser import ConfigParser

        parser = ConfigParser()
        parser.read(path)
        return parser
    elif ext == ".toml":
        import toml

        return toml.load(path)
    elif ext in (".yaml", ".yml"):
        import yaml

        with open(path, "r") as stream:
            return yaml.safe_load(stream)
    return dict()


def read_pyrekordbox_configuration():
    """Reads the pyrekordbox configuration.

    So far only the `pioneer-install-dir` and `pioneer-app-dir` fileds in the
    `rekordbox` section are supported. Supported configuration files are
    pyproject.toml, setup.cfg, rekordbox.toml, rekordbox.cfg and rekordbox.yml.
    The files are searched for in that order.

    Returns
    -------
    pyrekordbox_config : dict
        The pyrekordbox configuration data, if found.
    """
    files = ["pyproject.toml", "setup.cfg"]
    for ext in [".toml", ".cfg", ".yaml", ".yml"]:
        files.append("pyrekordbox" + ext)

    for file in files:
        try:
            data = _read_config_file(file)
            config = dict(data["rekordbox"])
            logger.debug("Read configuration from '%s'", file)
        except (ImportError, FileNotFoundError) as e:
            logger.debug("Could not read config file '%s': %s", file, e)
        except KeyError:
            pass
        else:
            if config:
                return config
    return dict()


def update_config(
    pioneer_install_dir: Union[str, Path] = None,
    pioneer_app_dir: Union[str, Path] = None,
    rb5_install_dirname: str = "",
    rb6_install_dirname: str = "",
    rb7_install_dirname: str = "",
):
    """Update the pyrekordbox configuration.

    This method scans the system for the Rekordbox installation and application data
    directories and extracts the reuired file locations. For this the default Pioneer
    directories (installation and application data) are used. If the method fails to
    find the directories they can be supplied as parameters. If no Rekordbox
    installation is found the fileds are left unchanged.
    On import configuration with the default locations is loaded.

    Parameters
    ----------
    pioneer_install_dir : str or Path, optional
        The path to the Pioneer installation directory. This is where the program files
        of Pioneer applications are stored. By default, the normal location of Pioneer
        programs is used.
    pioneer_app_dir : str or Path, optional
        The path to the Pioneer application directory. This is where the application
        user data of Pioneer programs is stored. By default, the normal location of
        the Pioneer application data is used.
    rb5_install_dirname : str, optional
        The name of the Rekordbox 5 installation directory. By default, the normal
        directory name is used (Windows: 'rekordbox 5.x.x', macOS: 'rekordbox 5.app').
    rb6_install_dirname : str, optional
        The name of the Rekordbox 6 installation directory. By default, the normal
        directory name is used (Windows: 'rekordbox 6.x.x', macOS: 'rekordbox 6.app').
    rb7_install_dirname : str, optional
        The name of the Rekordbox 7 installation directory. By default, the normal
        directory name is used (Windows: 'rekordbox 7.x.x', macOS: 'rekordbox 7.app').
    """
    # Read config file
    conf = read_pyrekordbox_configuration()
    if pioneer_install_dir is None and "pioneer-install-dir" in conf:
        pioneer_install_dir = conf["pioneer-install-dir"]
    if pioneer_app_dir is None and "pioneer-app-dir" in conf:
        pioneer_app_dir = conf["pioneer-app-dir"]
    if not rb5_install_dirname and "rekordbox5-install-dirname" in conf:
        rb5_install_dirname = conf["rekordbox5-install-dirname"]
    if not rb6_install_dirname and "rekordbox6-install-dirname" in conf:
        rb6_install_dirname = conf["rekordbox6-install-dirname"]
    if not rb7_install_dirname and "rekordbox7-install-dirname" in conf:
        rb7_install_dirname = conf["rekordbox7-install-dirname"]

    # Pioneer installation directory
    try:
        pioneer_install_dir = get_pioneer_install_dir(pioneer_install_dir)
        __config__["pioneer"]["install_dir"] = pioneer_install_dir
    except FileNotFoundError as e:
        logger.warning(e)
        return

    # Pioneer application data directory
    try:
        pioneer_app_dir = get_pioneer_app_dir(pioneer_app_dir)
        __config__["pioneer"]["app_dir"] = pioneer_app_dir
    except FileNotFoundError as e:
        logger.warning(e)
        return

    # Update Rekordbox 5 config
    try:
        conf = _get_rb5_config(pioneer_install_dir, pioneer_app_dir, rb5_install_dirname)
        __config__["rekordbox5"].update(conf)
    except FileNotFoundError as e:
        logger.info(e)

    # Update Rekordbox 6 config
    try:
        conf = _get_rb6_config(pioneer_install_dir, pioneer_app_dir, rb6_install_dirname)
        __config__["rekordbox6"].update(conf)
    except FileNotFoundError as e:
        logger.info(e)

    # Update Rekordbox 7 config
    try:
        conf = _get_rb7_config(pioneer_install_dir, pioneer_app_dir, rb7_install_dirname)
        __config__["rekordbox7"].update(conf)
    except FileNotFoundError as e:
        logger.info(e)


def get_config(section: str, key: str = None):
    """Gets a section or value of the pyrekordbox configuration.

    Parameters
    ----------
    section : str
        The name of the section.
    key : str, optional
        The name of the specific value to return. If not given all values of the section
        are returned as dictionary.

    Returns
    -------
    data : str or Path or dict
        The data of a section or a specific configuration value.
    """
    # Update config if not done yet
    if not __config__[section]:
        update_config()

    conf = __config__[section]
    if key is None:
        return conf
    return conf[key]


def pformat_config(indent: str = "   ", hw: int = 14, delim: str = " = ") -> str:
    """Returns a formatted string of the pyrekordbox configurations."""
    pioneer = get_config("pioneer")
    rb5 = get_config("rekordbox5")
    rb6 = get_config("rekordbox6")
    rb7 = get_config("rekordbox7")

    lines = ["Pioneer:"]
    lines += [f"{indent}{k + delim:<{hw}} {pioneer[k]}" for k in sorted(pioneer.keys())]
    lines.append("Rekordbox 5:")
    if rb5:
        lines += [f"{indent}{k + delim:<{hw}} {rb5[k]}" for k in sorted(rb5.keys())]
    lines.append("Rekordbox 6:")
    if rb6:
        rb6_keys = [k for k in rb6.keys() if k not in ("dp", "p")]
        lines += [f"{indent}{k + delim:<{hw}} {rb6[k]}" for k in sorted(rb6_keys)]
    lines.append("Rekordbox 7:")
    if rb7:
        rb7_keys = [k for k in rb7.keys() if k not in ("dp", "p")]
        lines += [f"{indent}{k + delim:<{hw}} {rb7[k]}" for k in sorted(rb7_keys)]
    return "\n".join(lines)


def show_config() -> None:
    """Prints a formatted string of the pyrekordbox configurations."""
    print(pformat_config())



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

import os

import numpy as np
import pytest
from numpy.testing import assert_equal

from pyrekordbox import anlz

TEST_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".testdata")
ANLZ_ROOT = os.path.join(TEST_ROOT, "export", "PIONEER", "USBANLZ")
ANLZ_DIRS = list(anlz.walk_anlz_paths(ANLZ_ROOT))
ANLZ_FILES = [paths for _, paths in ANLZ_DIRS]


def test_parse():
    for root, files in ANLZ_DIRS:
        for path in files.values():
            anlz.AnlzFile.parse_file(path)


def test_rebuild():
    for root, files in ANLZ_DIRS:
        for path in files.values():
            file = anlz.AnlzFile.parse_file(path)
            data = file.build()
            assert len(data) == file.file_header.len_file
            _ = anlz.AnlzFile.parse(data)


def test_read_anlz_files():
    for root, files in ANLZ_DIRS:
        anlz_files = anlz.read_anlz_files(root)
        assert len(files) == len(anlz_files)


# -- Tags ------------------------------------------------------------------------------


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_pqtz_tag_getters(paths):
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PQTZ")
    beats, bpms, times = tag.get()
    nbeats = len(beats)
    # Check that shape of arrays are equal
    assert nbeats == len(bpms)
    assert nbeats == len(times)
    # Check that the beats array only contains the values 1-4
    if nbeats:
        assert_equal(np.sort(np.unique(beats)), [1, 2, 3, 4])

    # Check other getters
    assert_equal(beats, tag.get_beats())
    assert_equal(bpms, tag.get_bpms())
    assert_equal(times, tag.get_times())


def test_pqtz_tag_set_beats():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PQTZ")

    beats = np.ones(tag.count)
    tag.set_beats(beats)
    assert_equal(tag.get_beats(), beats)


def test_pqtz_tag_set_bpms():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PQTZ")

    bpms = 100 * np.ones(tag.count, dtype=np.float64)
    tag.set_bpms(bpms)
    assert_equal(tag.get_bpms(), bpms)


def test_pqtz_tag_set_times():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PQTZ")

    times = 0.5 * np.arange(tag.count)
    tag.set_times(times)
    assert_equal(tag.get_times(), times)


def test_pqtz_tag_set():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PQTZ")

    beats = np.ones(tag.count)
    bpms = 100 * np.ones(tag.count, dtype=np.float64)
    times = 0.5 * np.arange(tag.count)
    tag.set(beats, bpms, times)
    assert_equal(tag.get_bpms(), bpms)
    assert_equal(tag.get_bpms(), bpms)
    assert_equal(tag.get_times(), times)


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_ppth_tag_getters(paths):
    tmp = ""
    for path in paths.values():
        file = anlz.AnlzFile.parse_file(path)
        tag = file.get_tag("PPTH")
        p = tag.get()
        if not tmp:
            tmp = p
        else:
            assert tmp == p


def test_ppth_tag_setters():
    paths = ANLZ_FILES[0]
    extected = r"C:/new/path/to/file.mp3"
    for path in paths.values():
        file = anlz.AnlzFile.parse_file(path)
        tag = file.get_tag("PPTH")
        tag.set(r"C:\new\path\to\file.mp3")
        assert tag.get() == extected


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_pwav_tag_getters(paths):
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PWAV")

    heights, color = tag.get()
    assert len(heights) == len(color)
    assert np.all(np.logical_and(0 <= heights, heights <= 31))
    assert np.all(np.logical_and(0 <= color, color <= 7))


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_pwv2_tag_getters(paths):
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    tag = file.get_tag("PWV2")

    heights, color = tag.get()
    assert len(heights) == len(color)
    assert np.all(np.logical_and(0 <= heights, heights <= 31))
    assert np.all(np.logical_and(0 <= color, color <= 7))


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_pwv3_tag_getters(paths):
    file = anlz.AnlzFile.parse_file(paths["EXT"])
    tag = file.get_tag("PWV3")

    heights, color = tag.get()
    assert len(heights) == len(color)
    assert np.all(np.logical_and(0 <= heights, heights <= 31))
    assert np.all(np.logical_and(0 <= color, color <= 7))


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_pwv4_tag_getters(paths):
    file = anlz.AnlzFile.parse_file(paths["EXT"])
    tag = file.get_tag("PWV4")

    heights, colors, blues = tag.get()
    assert len(heights) == len(colors)
    assert len(heights) == len(blues)


@pytest.mark.parametrize("paths", ANLZ_FILES)
def test_pwv5_tag_getters(paths):
    file = anlz.AnlzFile.parse_file(paths["EXT"])
    tag = file.get_tag("PWV5")

    heights, colors = tag.get()
    assert len(heights) == colors.shape[0]


# -- File ------------------------------------------------------------------------------


def test_anlzfile_getall_tags():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    key = "PPTH"
    tags = file.getall_tags(key)
    assert len(tags) == 1
    assert tags[0].get() == file.get(key)


def test_anlzfile_get():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    key = "PPTH"
    tag = file.get_tag(key)
    assert file.get(key) == tag.get()


def test_anlzfile_getall():
    paths = ANLZ_FILES[0]
    file = anlz.AnlzFile.parse_file(paths["DAT"])
    key = "PPTH"
    tag = file.get_tag(key)
    values = file.getall(key)
    assert len(values) == 1
    assert values[0] == tag.get()



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2022-04-10

from .anlz import AnlzFile, get_anlz_paths, read_anlz_files, walk_anlz_paths
from .config import get_config, show_config, update_config
from .db6 import Rekordbox6Database
from .logger import logger
from .mysettings import (
    DevSettingFile,
    DjmMySettingFile,
    MySetting2File,
    MySettingFile,
    get_mysetting_paths,
    read_mysetting_file,
)
from .rbxml import RekordboxXml, XmlAttributeKeyError, XmlDuplicateError

try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

import os
import re
import shutil
import tempfile
import urllib.request
from datetime import datetime, timezone
from pathlib import Path

import pytest
from pytest import mark
from sqlalchemy import text
from sqlalchemy.orm.query import Query

from pyrekordbox import Rekordbox6Database
from pyrekordbox.db6 import tables
from pyrekordbox.db6.smartlist import LogicalOperator, Operator, Property, SmartList
from pyrekordbox.db6.tables import datetime_to_str, string_to_datetime

TEST_ROOT = Path(__file__).parent.parent / ".testdata"
LOCKED = TEST_ROOT / "rekordbox 6" / "master_locked.db"
UNLOCKED = TEST_ROOT / "rekordbox 6" / "master_unlocked.db"
UNLOCKED_COPY = TEST_ROOT / "rekordbox 6" / "master_unlocked_copy.db"
UNLOCKED_OUT = TEST_ROOT / "rekordbox 6" / "master_unlocked_out.db"
MASTER_PLAYLIST_SRC = TEST_ROOT / "rekordbox 6" / "masterPlaylists6_template.xml"
MASTER_PLAYLIST_DST = TEST_ROOT / "rekordbox 6" / "masterPlaylists6.xml"
# Create a copy of the masterPlaylists6.xml file
shutil.copy(MASTER_PLAYLIST_SRC, MASTER_PLAYLIST_DST)

DB = Rekordbox6Database(UNLOCKED, unlock=False)

# Content IDs
CID1 = 178162577  # Demo Track 1
CID2 = 66382436  # Demo Track 2
CID3 = 181094952  # HORN
CID4 = 24401986  # NOISE

# Playlist ID
PID1 = 2602250856  # Trial playlist - Cloud Library Sync

KEY_SOURCES = [
    {
        "url": r"https://raw.githubusercontent.com/mganss/CueGen/19878e6eb3f586dee0eb3eb4f2ce3ef18309de9d/CueGen/Generator.cs",  # noqa: E501
        "regex": re.compile(
            r'((.|\n)*)Config\.UseSqlCipher.*\?.*"(?P<dp>.*)".*:.*null',
            flags=re.IGNORECASE | re.MULTILINE,
        ),
    },
    {
        "url": r"https://raw.githubusercontent.com/dvcrn/go-rekordbox/8be6191ba198ed7abd4ad6406d177ed7b4f749b5/cmd/getencryptionkey/main.go",  # noqa: E501
        "regex": re.compile(
            r'((.|\n)*)fmt\.Print\("(?P<dp>.*)"\)', flags=re.IGNORECASE | re.MULTILINE
        ),
    },
]


def download_db6_key():
    dp = ""
    for source in KEY_SOURCES:
        url = source["url"]
        regex = source["regex"]

        res = urllib.request.urlopen(url)
        data = res.read().decode("utf-8")
        match = regex.match(data)
        if match:
            dp = match.group("dp")
            break
    if dp:
        return dp
    else:
        raise Exception("No key found in the online sources.")


@pytest.fixture
def db():
    """Return a clean Rekordbox v6 database instance and close it after tests."""
    shutil.copy(UNLOCKED, UNLOCKED_COPY)
    shutil.copy(MASTER_PLAYLIST_SRC, MASTER_PLAYLIST_DST)
    db = Rekordbox6Database(UNLOCKED_COPY, unlock=False)
    yield db
    db.close()


def test_open_rekordbox_database():
    db = Rekordbox6Database(UNLOCKED, unlock=False)
    db.session.execute(text("SELECT name FROM sqlite_master WHERE type='table';"))
    db.close()


def test_unlock_rekordbox_database():
    key = download_db6_key()
    db = Rekordbox6Database(LOCKED, key=key, unlock=True)
    db.session.execute(text("SELECT name FROM sqlite_master WHERE type='table';"))
    db.close()


def test_close_open():
    db = Rekordbox6Database(UNLOCKED, unlock=False)
    db.close()
    db.open()
    _ = db.get_content()[0]  # Try to query the database
    db.close()


@mark.parametrize("dt", [datetime.now(), datetime.now(tz=timezone.utc)])
def test_datetime_to_string(dt):
    datetime_to_str(dt)


@mark.parametrize(
    "s",
    [
        "2025-04-12 19:11:29.274 +00:00",
        "2025-04-12 19:11:29.274 +01:00",
        "2025-04-12 19:11:29.274 -01:00",
        "2025-04-12 19:11:29.274 -05:00 (Central Daylight Time)",
    ],
)
def test_string_to_datetime(s):
    dt = string_to_datetime(s)
    # make sure no timezone is present
    assert dt.tzinfo is None


@mark.parametrize(
    "name,cls",
    [
        ("get_active_censor", tables.DjmdActiveCensor),
        ("get_album", tables.DjmdAlbum),
        ("get_artist", tables.DjmdArtist),
        ("get_category", tables.DjmdCategory),
        ("get_color", tables.DjmdColor),
        ("get_content", tables.DjmdContent),
        ("get_cue", tables.DjmdCue),
        ("get_device", tables.DjmdDevice),
        ("get_genre", tables.DjmdGenre),
        ("get_history", tables.DjmdHistory),
        ("get_hot_cue_banklist", tables.DjmdHotCueBanklist),
        ("get_key", tables.DjmdKey),
        ("get_label", tables.DjmdLabel),
        ("get_menu_items", tables.DjmdMenuItems),
        ("get_mixer_param", tables.DjmdMixerParam),
        ("get_my_tag", tables.DjmdMyTag),
        ("get_playlist", tables.DjmdPlaylist),
        ("get_property", tables.DjmdProperty),
        ("get_related_tracks", tables.DjmdRelatedTracks),
        ("get_sampler", tables.DjmdSampler),
        ("get_sort", tables.DjmdSort),
        ("get_agent_registry", tables.AgentRegistry),
        ("get_cloud_agent_registry", tables.CloudAgentRegistry),
        ("get_content_active_censor", tables.ContentActiveCensor),
        ("get_content_cue", tables.ContentCue),
        ("get_content_file", tables.ContentFile),
        ("get_hot_cue_banklist_cue", tables.HotCueBanklistCue),
        ("get_image_file", tables.ImageFile),
        ("get_setting_file", tables.SettingFile),
        ("get_uuid_map", tables.UuidIDMap),
    ],
)
def test_getter(name, cls):
    getter = getattr(DB, name)
    # Test return is query
    query = getter()
    assert isinstance(query, Query)
    assert query.column_descriptions[0]["type"] == cls

    # Test type of query result is the right table class
    res = query.first()
    assert res is None or isinstance(res, cls)


@mark.parametrize(
    "name,cls",
    [
        ("get_active_censor", tables.DjmdActiveCensor),
        ("get_album", tables.DjmdAlbum),
        ("get_artist", tables.DjmdArtist),
        ("get_category", tables.DjmdCategory),
        ("get_color", tables.DjmdColor),
        ("get_content", tables.DjmdContent),
        ("get_cue", tables.DjmdCue),
        ("get_device", tables.DjmdDevice),
        ("get_genre", tables.DjmdGenre),
        ("get_history", tables.DjmdHistory),
        ("get_hot_cue_banklist", tables.DjmdHotCueBanklist),
        ("get_key", tables.DjmdKey),
        ("get_label", tables.DjmdLabel),
        ("get_menu_items", tables.DjmdMenuItems),
        ("get_mixer_param", tables.DjmdMixerParam),
        ("get_my_tag", tables.DjmdMyTag),
        ("get_playlist", tables.DjmdPlaylist),
        # ("get_property", tables.DjmdProperty),
        ("get_related_tracks", tables.DjmdRelatedTracks),
        ("get_sampler", tables.DjmdSampler),
        ("get_sort", tables.DjmdSort),
        # ("get_agent_registry", tables.AgentRegistry),
        ("get_cloud_agent_registry", tables.CloudAgentRegistry),
        ("get_content_active_censor", tables.ContentActiveCensor),
        ("get_content_cue", tables.ContentCue),
        ("get_content_file", tables.ContentFile),
        ("get_hot_cue_banklist_cue", tables.HotCueBanklistCue),
        ("get_image_file", tables.ImageFile),
        ("get_setting_file", tables.SettingFile),
        ("get_uuid_map", tables.UuidIDMap),
    ],
)
def test_getter_by_id(name, cls):
    # Get method from class
    getter = getattr(DB, name)

    # Try to get a valid ID
    item = getter().first()
    if item is None:
        return
    id_ = item.ID

    # Test type of result is a table class, not the query
    res = getter(ID=id_)
    assert res is None or isinstance(res, cls)


@mark.parametrize(
    "parent_name,key,cls",
    [
        ("get_history", "HistoryID", tables.DjmdSongHistory),
        ("get_hot_cue_banklist", "HotCueBanklistID", tables.DjmdSongHotCueBanklist),
        ("get_my_tag", "MyTagID", tables.DjmdSongMyTag),
        ("get_playlist", "PlaylistID", tables.DjmdSongPlaylist),
        ("get_related_tracks", "RelatedTracksID", tables.DjmdSongPlaylist),
        ("get_sampler", "SamplerID", tables.DjmdSongPlaylist),
    ],
)
def test_songs_getters(parent_name, key, cls):
    # Get list (containing songs) getter
    getter = getattr(DB, parent_name)
    # Try to get a valid list ID
    query = getter()
    if not query.count():
        return  # No data to check...

    item = query.first()
    id_ = item.ID

    # Get list songs getter
    getter = getattr(DB, f"{parent_name}_songs")
    # Get song items
    query = getter(**{key: id_})
    if not query.count():
        return  # No data to check...

    assert isinstance(query.first(), cls)


def test_mixer_gain_setters(db):
    for item in db.get_mixer_param():
        # Check Gain setter
        low, high, value = int(item.GainLow), int(item.GainHigh), item.Gain
        item.Gain = value
        db.flush()
        assert item.GainLow == low
        assert item.GainHigh == high

        # Check Peak setter
        low, high, value = int(item.PeakLow), int(item.PeakHigh), item.Peak
        item.Peak = value
        db.flush()
        assert item.PeakLow == low
        assert item.PeakHigh == high


@mark.parametrize(
    "search,ids",
    [
        ("Demo Track 1", [CID1]),  # Title
        ("Demo Track 2", [CID2]),  # Title
        ("Loopmasters", [CID1, CID2]),  # Label/Artist Name
        ("Noise", [CID4]),  # lowercase
        ("NOIS", [CID4]),  # incomplete
    ],
)
def test_search_content(search, ids):
    results = DB.search_content(search)
    for id_, res in zip(ids, results):
        assert int(res.ID) == id_


def test_increment_local_usn(db):
    old = db.get_local_usn()
    db.increment_local_usn()
    assert db.get_local_usn() == old + 1

    db.increment_local_usn(1)
    assert db.get_local_usn() == old + 2

    db.increment_local_usn(2)
    assert db.get_local_usn() == old + 4

    with pytest.raises(ValueError):
        db.increment_local_usn(0)

    with pytest.raises(ValueError):
        db.increment_local_usn(-1)


def test_autoincrement_local_usn(db):
    old_usn = db.get_local_usn()  # store USN before changes
    track1 = db.get_content(ID=CID1)
    track2 = db.get_content(ID=CID2)
    track3 = db.get_content(ID=CID3)
    playlist = db.get_playlist(ID=PID1)
    with db.session.no_autoflush:
        # Change one field in first track (+1)
        track1.Title = "New title 1"
        # Change two fields in second track (+2)
        track2.Title = "New title 2"
        track2.BPM = 12900
        # Delete row from table )+1)
        db.delete(track3)
        # Change name of playlist (+1)
        playlist.Name = "New name"

        # Auto-increment USN
        new_usn = db.autoincrement_usn()

    # Check local Rekordbox USN and instance USN's
    assert new_usn == old_usn + 5
    assert track1.rb_local_usn == old_usn + 1
    assert track2.rb_local_usn == old_usn + 3
    # USN of deleted rows obviously don't get updated
    assert playlist.rb_local_usn == new_usn


def _check_playlist_xml(db):
    # Check that playlist is in XML and update time is correct
    for pl in db.get_playlist():
        plxml = db.playlist_xml.get(pl.ID)
        ts = plxml["Timestamp"]
        diff = pl.updated_at - ts
        if abs(diff.total_seconds()) > 1:
            return False
    return True


def _check_playlist_xml_delete(db):
    # Check that there are no items in the XML that are not in the db
    for plxml in db.playlist_xml.get_playlists():
        if plxml["Lib_Type"] != 0:
            continue
        pid = int(plxml["Id"], 16)
        if db.query(tables.DjmdPlaylist).filter_by(ID=pid).count() != 1:
            return False
    return True


def test_add_song_to_playlist(db):
    usn_old = db.get_local_usn()
    mtime_old = db.get_playlist(ID=PID1).updated_at

    # test adding song to playlist
    song = db.add_to_playlist(PID1, CID1)
    db.commit()

    pl = db.get_playlist(ID=PID1)
    assert len(pl.Songs) == 1
    assert pl.Songs[0].ContentID == str(CID1)
    assert song.TrackNo == 1

    # Test USN and update time are correct
    assert pl.updated_at == mtime_old
    assert song.rb_local_usn == usn_old + 1
    assert db.get_local_usn() == usn_old + 1

    # test raising error when adding song to playlist with wrong TrackNo
    with pytest.raises(ValueError):
        db.add_to_playlist(PID1, CID2, track_no=0)
    with pytest.raises(ValueError):
        db.add_to_playlist(PID1, CID2, track_no=3)

    assert _check_playlist_xml(db)


def test_add_song_to_playlist_trackno_end(db):
    old_usn = db.get_local_usn()
    song1 = db.add_to_playlist(PID1, CID1)
    song2 = db.add_to_playlist(PID1, CID2)
    song3 = db.add_to_playlist(PID1, CID3)
    db.commit()
    assert song1.TrackNo == 1
    assert song2.TrackNo == 2
    assert song3.TrackNo == 3
    assert db.get_local_usn() == old_usn + 3

    assert _check_playlist_xml(db)


def test_add_song_to_playlist_trackno_middle(db):
    song1 = db.add_to_playlist(PID1, CID1)
    song2 = db.add_to_playlist(PID1, CID2)
    assert song1.TrackNo == 1
    assert song2.TrackNo == 2
    db.commit()

    usn_old = db.get_local_usn()
    mtime_old = db.get_playlist(ID=PID1).updated_at

    # Insert song in the middle
    song3 = db.add_to_playlist(PID1, CID3, track_no=2)
    db.commit()
    assert song3.TrackNo == 2

    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: int(x.TrackNo))
    assert len(songs) == 3
    assert songs[0].ContentID == str(CID1)
    assert songs[0].TrackNo == 1
    assert songs[1].ContentID == str(CID3)
    assert songs[1].TrackNo == 2
    assert songs[2].ContentID == str(CID2)
    assert songs[2].TrackNo == 3

    # Test USN and update time are correct
    # First USN increment is for adding song, second for updating track no
    # of other songs in playlist
    assert pl.updated_at == mtime_old
    assert songs[1].rb_local_usn == usn_old + 2
    assert songs[2].rb_local_usn == usn_old + 2
    assert db.get_local_usn() == usn_old + 2

    assert _check_playlist_xml(db)


def test_remove_song_from_playlist_end(db):
    # Add songs to playlist
    db.add_to_playlist(PID1, CID1, track_no=1)
    db.add_to_playlist(PID1, CID2, track_no=2)
    song3 = db.add_to_playlist(PID1, CID3, track_no=3)
    sid3 = song3.ID
    db.commit()

    usn_old = db.get_local_usn()
    mtime_old = db.get_playlist(ID=PID1).updated_at

    # test removing song from playlist
    db.remove_from_playlist(PID1, sid3)
    db.commit()

    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: x.TrackNo)
    assert len(songs) == 2
    assert songs[0].ContentID == str(CID1)
    assert songs[0].TrackNo == 1
    assert songs[1].ContentID == str(CID2)
    assert songs[1].TrackNo == 2
    pl = db.get_playlist(ID=PID1)
    # Test USN and update time are correct
    assert pl.updated_at == mtime_old
    assert db.get_local_usn() == usn_old + 1

    assert _check_playlist_xml(db)


def test_remove_song_from_playlist_middle(db):
    # Add songs to playlist
    db.add_to_playlist(PID1, CID1, track_no=1)
    song2 = db.add_to_playlist(PID1, CID2, track_no=2)
    db.add_to_playlist(PID1, CID3, track_no=3)
    sid2 = song2.ID
    db.commit()
    usn_old = db.get_local_usn()

    # test removing song from playlist
    db.remove_from_playlist(PID1, sid2)
    db.commit()

    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: x.TrackNo)
    assert len(songs) == 2
    assert songs[0].ContentID == str(CID1)
    assert songs[0].TrackNo == 1
    assert songs[1].ContentID == str(CID3)
    assert songs[1].TrackNo == 2

    # Check USN is correct
    assert db.get_local_usn() == usn_old + 2
    assert songs[1].rb_local_usn == usn_old + 2

    assert _check_playlist_xml(db)


def test_move_in_playlist_forward(db):
    # Add songs to playlist
    s1 = db.add_to_playlist(PID1, CID1)
    s2 = db.add_to_playlist(PID1, CID2)
    s3 = db.add_to_playlist(PID1, CID3)
    s4 = db.add_to_playlist(PID1, CID4)
    db.commit()
    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: x.TrackNo)
    assert [int(s.ContentID) for s in songs] == [CID1, CID2, CID3, CID4]
    usn_old = db.get_local_usn()

    # Move song forward
    db.move_song_in_playlist(PID1, s3, 1)
    db.commit()
    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: x.TrackNo)
    assert [int(s.ContentID) for s in songs] == [CID3, CID1, CID2, CID4]

    # Check USN
    expected_usn = usn_old + 1
    assert db.get_local_usn() == expected_usn
    assert s1.rb_local_usn == expected_usn
    assert s2.rb_local_usn == expected_usn
    assert s3.rb_local_usn == expected_usn
    assert s4.rb_local_usn == usn_old

    assert _check_playlist_xml(db)


def test_move_in_playlist_backward(db):
    # Add songs to playlist
    s1 = db.add_to_playlist(PID1, CID1)
    s2 = db.add_to_playlist(PID1, CID2)
    s3 = db.add_to_playlist(PID1, CID3)
    s4 = db.add_to_playlist(PID1, CID4)
    db.commit()
    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: x.TrackNo)
    assert [int(s.ContentID) for s in songs] == [CID1, CID2, CID3, CID4]
    usn_old = db.get_local_usn()

    # Move song backward
    db.move_song_in_playlist(PID1, s1, 3)
    db.commit()
    pl = db.get_playlist(ID=PID1)
    songs = sorted(pl.Songs, key=lambda x: x.TrackNo)
    assert [int(s.ContentID) for s in songs] == [CID2, CID3, CID1, CID4]

    # Check USN
    expected_usn = usn_old + 1
    assert db.get_local_usn() == expected_usn
    assert s1.rb_local_usn == expected_usn
    assert s2.rb_local_usn == expected_usn
    assert s3.rb_local_usn == expected_usn
    assert s4.rb_local_usn == usn_old

    assert _check_playlist_xml(db)


def test_create_playlist(db):
    seqs = [pl.Seq for pl in db.get_playlist()]
    assert max(seqs) == 2
    old_usn = db.get_local_usn()

    # Create playlist
    pl = db.create_playlist("Test playlist")
    pid = pl.ID
    db.commit()

    # Check if playlist was created correctly
    pl = db.get_playlist(ID=pid)
    assert pl.Name == "Test playlist"
    assert pl.Seq == 3
    assert pl.Attribute == 0
    assert pl.Songs == []
    assert pl.ParentID == "root"
    assert pl.Children == []

    # Check USN is correct (+1 for creating, +1 for renaming)
    assert pl.rb_local_usn == old_usn + 2
    assert db.get_local_usn() == old_usn + 2

    # Try to add song to playlist
    db.add_to_playlist(pl, CID1)
    db.commit()

    pl = db.get_playlist(ID=pid)
    assert len(pl.Songs) == 1

    # Check if playlist was added to xml
    plxml = db.playlist_xml.get(pl.ID)
    assert plxml is not None
    assert _check_playlist_xml(db)


def test_create_playlist_seq_middle(db):
    seqs = [pl.Seq for pl in db.get_playlist()]
    assert max(seqs) == 2
    old_usn = db.get_local_usn()

    # Create playlist
    pl = db.create_playlist("playlist1")
    pid1 = pl.ID
    assert pl.Seq == 3
    db.commit()
    # Check USN is correct (+1 for creating, +1 for renaming)
    assert pl.rb_local_usn == old_usn + 2
    assert db.get_local_usn() == old_usn + 2

    old_usn = db.get_local_usn()
    pl = db.create_playlist("playlist2", seq=3)
    pid2 = pl.ID
    db.commit()

    pl1 = db.get_playlist(ID=pid1)
    pl2 = db.get_playlist(ID=pid2)
    assert pl1.Seq == 4
    assert pl2.Seq == 3
    # Check USN is correct
    assert db.get_local_usn() == old_usn + 3  # +2 for creating, +1 for moving others
    assert pl1.rb_local_usn == old_usn + 1
    assert pl2.rb_local_usn == old_usn + 3

    assert _check_playlist_xml(db)


def test_create_playlist_folder(db):
    seqs = [pl.Seq for pl in db.get_playlist()]
    assert max(seqs) == 2
    usn_old = db.get_local_usn()

    # Create playlist
    pl = db.create_playlist_folder("Test playlist folder")
    pid = pl.ID
    db.commit()

    # Check if playlist folder was created correctly
    pl = db.get_playlist(ID=pid)
    assert pl.Name == "Test playlist folder"
    assert pl.Seq == 3
    assert pl.Attribute == 1
    assert pl.Songs == []
    assert pl.ParentID == "root"
    assert pl.Children == []
    # Check USN is correct (+1 for creating, +1 for renaming)
    assert db.get_local_usn() == usn_old + 2
    assert pl.rb_local_usn == usn_old + 2

    # Try to add sub-playlist to playlist folder
    db.create_playlist("Test playlist", parent=pl)
    db.commit()

    pl = db.get_playlist(ID=pid)
    assert len(pl.Children) == 1

    assert _check_playlist_xml(db)


def test_create_smart_playlist(db):
    seqs = [pl.Seq for pl in db.get_playlist()]
    assert max(seqs) == 2
    old_usn = db.get_local_usn()

    # Create smart list
    smart = SmartList(LogicalOperator.ALL)
    smart.add_condition(Property.ARTIST, Operator.EQUAL, "Loopmasters")

    # create smart playlist from smart list
    pl = db.create_smart_playlist("Smart playlist", smart)
    pid = pl.ID
    db.commit()

    # Check if playlist was created correctly
    pl = db.get_playlist(ID=pid)
    assert pl.Name == "Smart playlist"
    assert pl.Seq == 3
    assert pl.Attribute == 4
    assert pl.Songs == []
    assert pl.ParentID == "root"
    assert pl.Children == []

    # Check USN is correct (+1 for creating, +1 for renaming)
    assert pl.rb_local_usn == old_usn + 2
    assert db.get_local_usn() == old_usn + 2

    # Check if playlist was added to xml
    plxml = db.playlist_xml.get(pl.ID)
    assert plxml is not None
    assert _check_playlist_xml(db)


def test_delete_playlist_empty_end(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    pl1 = db.create_playlist("sub playlist 1", parent=folder.ID)
    pl2 = db.create_playlist("sub playlist 2", parent=folder.ID)
    pl3 = db.create_playlist("sub playlist 3", parent=folder.ID)
    assert pl1.Seq == 1 and pl2.Seq == 2 and pl3.Seq == 3
    db.commit()
    usn_old = db.get_local_usn()

    # Delete playlist
    pl = db.get_playlist(Name="sub playlist 3").one()
    pid = pl.ID
    db.delete_playlist(pl)
    db.commit()

    # Check if playlist was deleted
    pl = db.get_playlist(ID=pid)
    assert pl is None
    # Check if playlist was deleted from xml
    plxml = db.playlist_xml.get(pid)
    assert plxml is None
    # Check USN is correct (+1 for deleting)
    assert db.get_local_usn() == usn_old + 1

    assert _check_playlist_xml(db)
    assert _check_playlist_xml_delete(db)


def test_delete_playlist_empty(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    pl1 = db.create_playlist("sub playlist 1", parent=folder.ID)
    pl2 = db.create_playlist("sub playlist 2", parent=folder.ID)
    pl3 = db.create_playlist("sub playlist 3", parent=folder.ID)
    assert pl1.Seq == 1 and pl2.Seq == 2 and pl3.Seq == 3
    db.commit()
    usn_old = db.get_local_usn()

    # Delete playlist
    pl = db.get_playlist(Name="sub playlist 2").one()
    pid = pl.ID
    db.delete_playlist(pl)
    db.commit()

    # Check if playlist was deleted
    pl = db.get_playlist(ID=pid)
    assert pl is None
    # Check if playlist was deleted from xml
    plxml = db.playlist_xml.get(pid)
    assert plxml is None
    # Check USN is correct (+1 for deleting, all moved playlists get same USN)
    assert db.get_local_usn() == usn_old + 1
    # Check if seq numbers in parent were updated
    pl = db.get_playlist(Name="sub playlist 3").one()
    assert pl.Seq == 2
    assert pl.rb_local_usn == usn_old + 1

    assert _check_playlist_xml(db)
    assert _check_playlist_xml_delete(db)


def test_delete_playlist_folder_empty(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    pl1 = db.create_playlist("sub playlist 1", parent=folder.ID)
    pl2 = db.create_playlist_folder("sub playlist folder", parent=folder.ID)
    pl3 = db.create_playlist("sub playlist 3", parent=folder.ID)
    assert pl1.Seq == 1 and pl2.Seq == 2 and pl3.Seq == 3
    db.commit()
    usn_old = db.get_local_usn()

    # Delete playlist
    pl = db.get_playlist(Name="sub playlist folder").one()
    pid = pl.ID
    db.delete_playlist(pl)
    db.commit()

    # Check if playlist was deleted
    pl = db.get_playlist(ID=pid)
    assert pl is None
    # Check if playlist was deleted from xml
    plxml = db.playlist_xml.get(pid)
    assert plxml is None
    # Check if seq numbers in parent were updated
    pl = db.get_playlist(Name="sub playlist 3").one()
    assert pl.Seq == 2
    # Check USN is correct (+1 for deleting)
    assert db.get_local_usn() == usn_old + 1
    assert pl.rb_local_usn == usn_old + 1

    assert _check_playlist_xml(db)
    assert _check_playlist_xml_delete(db)


def test_delete_playlist_non_empty(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    pl1 = db.create_playlist("sub playlist 1", parent=folder.ID)
    pl2 = db.create_playlist("sub playlist 2", parent=folder.ID)
    pl3 = db.create_playlist("sub playlist 3", parent=folder.ID)
    # Add songs to playlists
    sid1 = db.add_to_playlist(pl1, CID1).ID
    sid2 = db.add_to_playlist(pl2, CID2).ID
    sid3 = db.add_to_playlist(pl2, CID3).ID
    sid4 = db.add_to_playlist(pl3, CID4).ID
    db.commit()
    usn_old = db.get_local_usn()

    assert db.query(tables.DjmdSongPlaylist).count() == 4

    # Delete playlist
    pl = db.get_playlist(Name="sub playlist 2").one()
    pid = pl.ID
    db.delete_playlist(pl)
    db.commit()

    # Check if playlist was deleted
    pl = db.get_playlist(ID=pid)
    assert pl is None
    # Check if songs in playlist were deleted
    assert db.query(tables.DjmdSongPlaylist).count() == 2
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid1).count() == 1
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid2).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid3).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid4).count() == 1
    # Check if USN is correct (+1 for deleting with contents)
    assert db.get_local_usn() == usn_old + 1

    assert _check_playlist_xml(db)
    assert _check_playlist_xml_delete(db)


def test_delete_playlist_folder_non_empty(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    pl1 = db.create_playlist("sub playlist 1", parent=folder.ID)
    folder2 = db.create_playlist_folder("sub playlist folder", parent=folder.ID)
    pl2 = db.create_playlist("sub sub playlist", parent=folder2.ID)
    pl3 = db.create_playlist("sub playlist 3", parent=folder.ID)
    # Add songs to playlists
    sid1 = db.add_to_playlist(pl1, CID1).ID
    sid2 = db.add_to_playlist(pl2, CID2).ID
    sid3 = db.add_to_playlist(pl2, CID3).ID
    sid4 = db.add_to_playlist(pl3, CID4).ID
    db.commit()
    usn_old = db.get_local_usn()

    assert db.query(tables.DjmdSongPlaylist).count() == 4

    # Delete playlist
    pl = db.get_playlist(Name="sub playlist folder").one()
    pid = pl.ID

    pid2 = db.get_playlist(Name="sub sub playlist").one().ID

    db.delete_playlist(pl)
    db.commit()

    # Check if playlists were deleted
    pl = db.get_playlist(ID=pid)
    assert pl is None
    pl = db.get_playlist(ID=pid2)
    assert pl is None
    # Check if songs in playlist were deleted
    assert db.query(tables.DjmdSongPlaylist).count() == 2
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid1).count() == 1
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid2).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid3).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid4).count() == 1

    # Check if USN is correct (+1 for deleting with Seq update, +1 for children)
    assert db.get_local_usn() == usn_old + 2

    assert _check_playlist_xml(db)
    assert _check_playlist_xml_delete(db)


def test_delete_playlist_folder_chained(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    subfolder1 = db.create_playlist_folder("subfolder", parent=folder.ID)
    subfolder2 = db.create_playlist_folder("subsubfolder", parent=subfolder1.ID)
    subfolder3 = db.create_playlist_folder("subsubsubfolder", parent=subfolder2.ID)
    pl1 = db.create_playlist("sub playlist 1", parent=subfolder2.ID)
    pl2 = db.create_playlist("sub playlist 2", parent=subfolder3.ID)

    pid1 = folder.ID
    pid2 = subfolder1.ID
    pid3 = subfolder2.ID
    pid4 = subfolder3.ID
    pid5 = pl1.ID
    pid6 = pl2.ID

    # Add songs to playlists
    sid1 = db.add_to_playlist(pl1, CID1).ID
    sid2 = db.add_to_playlist(pl1, CID2).ID
    sid3 = db.add_to_playlist(pl2, CID3).ID
    sid4 = db.add_to_playlist(pl2, CID4).ID

    db.commit()
    usn_old = db.get_local_usn()

    db.delete_playlist(folder)
    db.commit()

    # Check if all playlists and songs were deleted
    assert db.get_playlist(ID=pid1) is None
    assert db.get_playlist(ID=pid2) is None
    assert db.get_playlist(ID=pid3) is None
    assert db.get_playlist(ID=pid5) is None
    assert db.get_playlist(ID=pid4) is None
    assert db.get_playlist(ID=pid6) is None
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid1).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid2).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid3).count() == 0
    assert db.query(tables.DjmdSongPlaylist).filter_by(ID=sid4).count() == 0

    # Check all the corresponding xml entries were deleted
    assert db.playlist_xml.get(pid1) is None
    assert db.playlist_xml.get(pid2) is None
    assert db.playlist_xml.get(pid3) is None
    assert db.playlist_xml.get(pid4) is None
    assert db.playlist_xml.get(pid5) is None
    assert db.playlist_xml.get(pid6) is None

    # Check if USN is correct (+1 for deleting with Seq update, +1 for children)
    assert db.get_local_usn() == usn_old + 2

    assert _check_playlist_xml(db)
    assert _check_playlist_xml_delete(db)


def test_move_playlist_seq(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    f1 = db.create_playlist_folder("f 1", parent=folder)
    _ = db.create_playlist_folder("f 2", parent=folder)
    db.create_playlist("pl 1", parent=folder)
    db.create_playlist("pl 2", parent=folder)
    db.create_playlist("pl 3", parent=folder)
    db.create_playlist("pl 4", parent=folder)
    db.create_playlist("sub pl 1", parent=f1)
    db.create_playlist("sub pl 2", parent=f1)
    db.create_playlist("sub pl 3", parent=f1)
    db.create_playlist("sub pl 4", parent=f1)
    db.commit()

    playlists = db.get_playlist(ParentID=folder.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["f 1", "f 2", "pl 1", "pl 2", "pl 3", "pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5, 6]

    playlists = db.get_playlist(ParentID=f1.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["sub pl 1", "sub pl 2", "sub pl 3", "sub pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4]

    usn_old = db.get_local_usn()
    # Move playlist 1 to position 5
    pl = db.get_playlist(Name="pl 1").one()
    mtime_old = pl.updated_at
    db.move_playlist(pl, seq=5)
    db.commit()

    playlists = db.get_playlist(ParentID=folder.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["f 1", "f 2", "pl 2", "pl 3", "pl 1", "pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5, 6]
    # Check that the usn and update time was updated correctly
    # First the moved playlist USN is set, then the other updated playlists
    # in the seq order
    pl1 = db.get_playlist(Name="pl 1").one()
    pl2 = db.get_playlist(Name="pl 2").one()
    pl3 = db.get_playlist(Name="pl 3").one()
    assert db.get_local_usn() == usn_old + 3
    assert pl1.rb_local_usn == usn_old + 1
    assert pl2.rb_local_usn == usn_old + 2
    assert pl3.rb_local_usn == usn_old + 3
    assert pl1.updated_at >= mtime_old
    assert pl2.updated_at >= mtime_old
    assert pl3.updated_at >= mtime_old

    usn_old = db.get_local_usn()
    # Move playlist 3 to position 2
    pl = db.get_playlist(Name="pl 3").one()
    mtime_old = pl.updated_at
    db.move_playlist(pl, seq=2)
    db.commit()

    playlists = db.get_playlist(ParentID=folder.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["f 1", "pl 3", "f 2", "pl 2", "pl 1", "pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5, 6]
    # Check that the usn and update time was updated correctly
    # First the moved playlist USN is set, then the other updated playlists
    # in the seq order
    pl2 = db.get_playlist(Name="pl 2").one()
    pl3 = db.get_playlist(Name="pl 3").one()
    f2 = db.get_playlist(Name="f 2").one()
    assert db.get_local_usn() == usn_old + 3
    assert pl3.rb_local_usn == usn_old + 1
    assert f2.rb_local_usn == usn_old + 2
    assert pl2.rb_local_usn == usn_old + 3
    assert pl2.updated_at >= mtime_old
    assert pl3.updated_at >= mtime_old
    assert f2.updated_at >= mtime_old

    assert _check_playlist_xml(db)


def test_move_playlist_parent(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    f1 = db.create_playlist_folder("f 1", parent=folder)
    _ = db.create_playlist_folder("f 2", parent=folder)
    db.create_playlist("pl 1", parent=folder)
    db.create_playlist("pl 2", parent=folder)
    db.create_playlist("pl 3", parent=folder)
    db.create_playlist("pl 4", parent=folder)
    db.create_playlist("sub pl 1", parent=f1)
    db.create_playlist("sub pl 2", parent=f1)
    db.create_playlist("sub pl 3", parent=f1)
    db.create_playlist("sub pl 4", parent=f1)
    db.commit()

    playlists = db.get_playlist(ParentID=folder.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["f 1", "f 2", "pl 1", "pl 2", "pl 3", "pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5, 6]

    playlists = db.get_playlist(ParentID=f1.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["sub pl 1", "sub pl 2", "sub pl 3", "sub pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4]

    # Move playlist 1 to sub playlist (at the end)
    old_usn = db.get_local_usn()
    pl = db.get_playlist(Name="pl 1").one()
    old_mtime = pl.updated_at
    db.move_playlist(pl, parent=f1)
    db.commit()

    playlists = db.get_playlist(ParentID=folder.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["f 1", "f 2", "pl 2", "pl 3", "pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5]
    playlists = db.get_playlist(ParentID=f1.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["sub pl 1", "sub pl 2", "sub pl 3", "sub pl 4", "pl 1"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5]

    # Check that the usn and update time was updated correctly
    pl = db.get_playlist(Name="pl 1").one()
    assert db.get_local_usn() == old_usn + 1
    assert pl.rb_local_usn == old_usn + 1
    assert pl.updated_at >= old_mtime

    # Move playlist 2 to sub playlist
    old_usn = db.get_local_usn()
    pl = db.get_playlist(Name="pl 2").one()
    old_mtime = pl.updated_at
    db.move_playlist(pl, seq=2, parent=f1)
    db.commit()

    playlists = db.get_playlist(ParentID=folder.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["f 1", "f 2", "pl 3", "pl 4"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4]
    playlists = db.get_playlist(ParentID=f1.ID).order_by(tables.DjmdPlaylist.Seq)
    expected = ["sub pl 1", "pl 2", "sub pl 2", "sub pl 3", "sub pl 4", "pl 1"]
    assert [p.Name for p in playlists] == expected
    assert [pl.Seq for pl in playlists] == [1, 2, 3, 4, 5, 6]
    # Check that the usn and update time was updated correctly
    _ = db.get_playlist(Name="sub pl 1").one()
    subpl2 = db.get_playlist(Name="sub pl 2").one()
    subpl3 = db.get_playlist(Name="sub pl 3").one()
    subpl4 = db.get_playlist(Name="sub pl 4").one()
    pl1 = db.get_playlist(Name="pl 1").one()
    pl = db.get_playlist(Name="pl 2").one()
    assert db.get_local_usn() == old_usn + 5
    assert pl.rb_local_usn == old_usn + 1
    assert pl.updated_at >= old_mtime
    assert subpl2.rb_local_usn == old_usn + 2
    assert subpl3.rb_local_usn == old_usn + 3
    assert subpl4.rb_local_usn == old_usn + 4
    assert pl1.rb_local_usn == old_usn + 5

    assert _check_playlist_xml(db)


def test_rename_playlist(db):
    # Create playlist structure
    folder = db.create_playlist_folder("folder")
    db.create_playlist("pl 1", parent=folder)
    db.create_playlist("pl 2", parent=folder)
    db.commit()

    pl = db.get_playlist(Name="pl 1").one()
    pid = pl.ID
    mtime_old = pl.updated_at
    usn_old = db.get_local_usn()

    # Rename playlist
    db.rename_playlist(pl, "pl 1 new")
    db.commit()

    pl = db.get_playlist(ID=pid)
    assert pl.Name == "pl 1 new"
    assert pl.updated_at >= mtime_old
    assert db.get_local_usn() == usn_old + 1
    assert pl.rb_local_usn == usn_old + 1

    assert _check_playlist_xml(db)


def test_get_playlist_contents(db):
    # Create playlist and add content
    pl = db.create_playlist("Test playlist")
    db.add_to_playlist(pl, CID1)
    db.add_to_playlist(pl, CID2)
    db.commit()

    # Get playlist contents
    contents = db.get_playlist_contents(pl).all()
    assert len(contents) == 2
    assert {c.ID for c in contents} == {str(CID1), str(CID2)}


def test_get_playlist_contents_smart(db):
    # Singe condition
    smart = SmartList(LogicalOperator.ALL)
    smart.add_condition(Property.ARTIST, Operator.EQUAL, "Loopmasters")
    pl = db.create_smart_playlist("Smart playlist", smart)
    db.commit()
    contents = db.get_playlist_contents(pl).all()
    assert len(contents) == 2
    assert {c.ID for c in contents} == {str(CID1), str(CID2)}

    # All conditions
    smart = SmartList(LogicalOperator.ALL)
    smart.add_condition(Property.ARTIST, Operator.EQUAL, "Loopmasters")
    smart.add_condition(Property.NAME, Operator.EQUAL, "Demo Track 1")
    pl = db.create_smart_playlist("Smart playlist", smart)
    db.commit()
    contents = db.get_playlist_contents(pl).all()
    assert len(contents) == 1
    assert {c.ID for c in contents} == {str(CID1)}

    # Any conditions
    smart = SmartList(LogicalOperator.ANY)
    smart.add_condition(Property.ARTIST, Operator.EQUAL, "Loopmasters")
    smart.add_condition(Property.NAME, Operator.EQUAL, "HORN")
    pl = db.create_smart_playlist("Smart playlist", smart)
    db.commit()
    contents = db.get_playlist_contents(pl).all()
    assert len(contents) == 3
    assert {c.ID for c in contents} == {str(CID1), str(CID2), str(CID3)}


def test_add_album(db):
    old_usn = db.get_local_usn()
    name = "test"
    db.add_album(name)
    db.commit()

    # Check that album was created and USN is incremented
    instance = db.get_album(Name=name).one()
    assert instance.Name == name
    assert instance.rb_local_usn == old_usn + 1
    assert db.get_local_usn() == old_usn + 1

    # Fail if album with same name is added
    with pytest.raises(ValueError):
        db.add_album(name)

    # Add album with album artist by artist
    artist = db.get_artist().first()
    album = db.add_album("album 2", artist=artist)
    assert album.AlbumArtistID == artist.ID

    # Add album with album artist by ID
    artist = db.get_artist().first()
    album = db.add_album("album 3", artist=artist.ID)
    assert album.AlbumArtistID == artist.ID


def test_add_artist(db):
    old_usn = db.get_local_usn()
    name = "test"
    db.add_artist(name)
    db.commit()

    # Check that album was created and USN is incremented
    instance = db.get_artist(Name=name).one()
    assert instance.Name == name
    assert instance.rb_local_usn == old_usn + 1
    assert db.get_local_usn() == old_usn + 1

    # Fail if album with same name is added
    with pytest.raises(ValueError):
        db.add_artist(name)


def test_add_genre(db):
    old_usn = db.get_local_usn()
    name = "test"
    db.add_genre(name)
    db.commit()

    # Check that album was created and USN is incremented
    instance = db.get_genre(Name=name).one()
    assert instance.Name == name
    assert instance.rb_local_usn == old_usn + 1
    assert db.get_local_usn() == old_usn + 1

    # Fail if album with same name is added
    with pytest.raises(ValueError):
        db.add_genre(name)


def test_add_label(db):
    old_usn = db.get_local_usn()
    name = "test"
    db.add_label(name)
    db.commit()

    # Check that album was created and USN is incremented
    instance = db.get_label(Name=name).one()
    assert instance.Name == name
    assert instance.rb_local_usn == old_usn + 1
    assert db.get_local_usn() == old_usn + 1

    # Fail if album with same name is added
    with pytest.raises(ValueError):
        db.add_label(name)


def test_add_content(db):
    path = os.path.join(TEST_ROOT, "empty.mp3")
    # Add content
    content = db.add_content(path, Title="Test")
    assert content.Title == "Test"
    assert content.FolderPath == path

    # Fail on duplicate content
    with pytest.raises(ValueError):
        db.add_content(path)


def test_get_anlz_paths():
    content = DB.get_content().first()

    anlz_dir = str(DB.get_anlz_dir(content)).replace("\\", "/")
    expected = r"share/PIONEER/USBANLZ/735/e8b81-e69b-41ad-80f8-9c0d7613b96d"
    assert anlz_dir.endswith(expected)


def test_to_json():
    # Check if saving to json works

    tmp = tempfile.NamedTemporaryFile(delete=False)
    try:
        DB.to_json(tmp.name)
    finally:
        tmp.close()
        os.remove(tmp.name)


def test_copy_unlocked():
    db = Rekordbox6Database(UNLOCKED, unlock=False)
    db.copy_unlocked(UNLOCKED_OUT)

    db2 = Rekordbox6Database(UNLOCKED_OUT, unlock=False)
    # Check everything got copied
    for name in tables.TABLES:
        table = getattr(tables, name)
        for row in db.query(table):
            data = row.to_dict()
            if name == "AgentRegistry":
                query = db2.query(table).filter_by(registry_id=row.registry_id)
            elif name == "DjmdProperty":
                query = db2.query(table).filter_by(DBID=row.DBID)
            else:
                query = db2.query(table).filter_by(ID=row.ID)
            data2 = query.one().to_dict()
            assert data == data2



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

import logging
from collections import abc
from pathlib import Path
from typing import Union

from construct import Int16ub

from . import structs
from .tags import TAGS

logger = logging.getLogger(__name__)

XOR_MASK = bytearray.fromhex("CB E1 EE FA E5 EE AD EE E9 D2 E9 EB E1 E9 F3 E8 E9 F4 E1")


class BuildFileLengthError(Exception):
    def __init__(self, struct, len_data):
        super().__init__(
            f"`len_file` ({struct.len_file}) of '{struct.type}' does not "
            f"match the data-length ({len_data})!"
        )


class AnlzFile(abc.Mapping):
    """Rekordbox `ANLZnnnn.xxx` binary file handler."""

    def __init__(self):
        self._path = ""
        self.file_header = None
        self.tags = list()

    @property
    def num_tags(self):
        return len(self.tags)

    @property
    def tag_types(self):
        return [tag.type for tag in self.tags]

    @property
    def path(self):
        return self._path

    @classmethod
    def parse(cls, data: bytes):
        """Parses the in-memory data of a Rekordbox analysis binary file.

        Parameters
        ----------
        data : bytes
            The in-memory binary contents of a Rekordbox analysis file.

        Returns
        -------
        self : AnlzFile
            The new instance with the parsed file content.
        """
        self = cls()
        self._parse(data)
        return self

    @classmethod
    def parse_file(cls, path: Union[str, Path]):
        """Reads and parses a Rekordbox analysis binary file.

        Parameters
        ----------
        path : str or Path
            The path of a Rekordbox analysis file which is used to read
            the file contents before parsing the binary data.

        Returns
        -------
        self : AnlzFile
            The new instance with the parsed file content.

        See Also
        --------
        AnlzFile.parse: Parses the data of a Rekordbox analysis file.
        """
        path = Path(path)
        ext = path.suffix.upper()
        if ext not in (".DAT", ".EXT", ".2EX"):
            raise ValueError(f"File type '{ext}' not supported!")

        logger.debug(f"Reading file {path.name}")
        with open(path, "rb") as fh:
            data = fh.read()

        self = cls.parse(data)
        self._path = path
        return self

    def _parse(self, data: bytes):
        file_header = structs.AnlzFileHeader.parse(data)
        tag_type = file_header.type
        assert tag_type == "PMAI"

        tags = list()
        i = file_header.len_header
        while i < file_header.len_file:
            # Get data starting from struct
            tag_data = data[i:]
            # Get the four byte struct type
            tag_type = tag_data[:4].decode("ascii")

            if tag_type == "PSSI":
                # The version that rekordbox 6 *exports* is garbled with an XOR mask.
                # All bytes after byte 17 (len_e) are XOR-masked with a pattern that is
                # generated by adding the value of len_e to each byte of XOR_MASK

                # Check if the file is garbled (only on exported files)
                # For this we check the validity of mood and bank
                # Mood: High=1, Mid=2, Low=3
                # Bank: 0-8
                mood = Int16ub.parse(tag_data[18:20])
                bank = Int16ub.parse(tag_data[28:30])
                if 1 <= mood <= 3 and 0 <= bank <= 8:
                    logger.debug("PSSI is not garbled!")
                else:
                    logger.debug("PSSI is garbled!")
                    # deobfuscate tag_data[18:] using xor with XOR_MASK+len_entries
                    len_entries = Int16ub.parse(tag_data[16:])
                    tag_data = bytearray(data[i : i + len(tag_data)])
                    for x in range(len(tag_data[18:])):
                        mask = XOR_MASK[x % len(XOR_MASK)] + len_entries
                        if mask > 255:
                            mask -= 256
                        tag_data[x + 18] ^= mask
                    tag_data = bytes(tag_data)

            try:
                # Parse the struct
                tag = TAGS[tag_type](tag_data)
                tags.append(tag)
                len_header = tag.struct.len_header
                len_tag = tag.struct.len_tag
                logger.debug(
                    "Parsed struct '%s' (len_header=%s, len_tag=%s)",
                    tag_type,
                    len_header,
                    len_tag,
                )
            except KeyError:
                logger.warning("Tag '%s' not supported!", tag_type)
                tag = structs.AnlzTag.parse(tag_data)
                len_tag = tag.len_tag
            i += len_tag

        self.file_header = file_header
        self.tags = tags

    def update_len(self):
        # Update struct lengths
        tags_len = 0
        for tag in self.tags:
            tag.update_len()
            tags_len += tag.struct.len_tag
        # Update file length
        len_file = self.file_header.len_header + tags_len
        self.file_header.len_file = len_file

    def build(self):
        self.update_len()
        header_data = structs.AnlzFileHeader.build(self.file_header)
        section_data = b"".join(tag.build() for tag in self.tags)
        data = header_data + section_data
        # Check `len_file`
        len_file = self.file_header.len_file
        len_data = len(data)
        if len_file != len_data:
            raise BuildFileLengthError(self.file_header, len_file)

        return data

    def save(self, path=""):
        path = path or self._path

        data = self.build()
        with open(path, "wb") as fh:
            fh.write(data)

    def get_tag(self, key):
        return self.__getitem__(key)[0]

    def getall_tags(self, key):
        return self.__getitem__(key)

    def get(self, key):
        return self.__getitem__(key)[0].get()

    def getall(self, key):
        return [tag.get() for tag in self.__getitem__(key)]

    def __len__(self):
        return len(self.keys())

    def __iter__(self):
        return iter(set(tag.type for tag in self.tags))

    def __getitem__(self, item):
        if item.isupper() and len(item) == 4:
            return [tag for tag in self.tags if tag.type == item]
        else:
            return [tag for tag in self.tags if tag.name == item]

    def __contains__(self, item):
        if item.isupper() and len(item) == 4:
            for tag in self.tags:
                if item == tag.type:
                    return True
        else:
            for tag in self.tags:
                if item == tag.name:
                    return True
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}({self.tag_types})"

    def set_path(self, path):
        tag = self.get_tag("PPTH")
        tag.set(path)



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2023-02-01

"""Binary structures of Rekordbox MySettings files.

Files in this format include:

- `DEVSETTINGS.DAT`
- `DJMMYSETTING.DAT`
- `MYSETTING.DAT`
- `MYSETTING2.DAT`

References
----------
.. [1] Jan Holthuis. rekordcrate.
   https://holzhaus.github.io/rekordcrate/rekordcrate/setting/struct.Setting.html
"""

from construct import Int8ul, Int16ul, Int32ul, PaddedString, Const, Padding
from construct import Enum, Bytes, Struct, Default, Select


# -- DjmSettings -----------------------------------------------------------------------

ChannelFaderCurve = Enum(Int8ul, steep_top=0x80, linear=0x81, steep_bottom=0x82)
CrossfaderCurve = Enum(Int8ul, constant=0x80, slow_cut=0x81, fast_cut=0x82)
HeadphonesPreEQ = Enum(Int8ul, post_eq=0x80, pre_eq=0x81)
HeadphonesMonoSplit = Enum(Int8ul, stereo=0x80, mono_split=0x81)
BeatFXQuantize = Enum(Int8ul, off=0x80, on=0x81)
MicLowCut = Enum(Int8ul, off=0x80, on=0x81)
TalkOverMode = Enum(Int8ul, advanced=0x80, normal=0x81)
TalkOverLevel = Enum(
    Int8ul,
    minus_24db=0x80,
    minus_18db=0x81,
    minus_12db=0x82,
    minus_6db=0x83
)
MidiChannel = Enum(
    Int8ul,
    one=0x80,
    two=0x81,
    three=0x82,
    four=0x83,
    five=0x84,
    six=0x85,
    seven=0x86,
    eight=0x87,
    nine=0x88,
    ten=0x89,
    eleven=0x8A,
    twelve=0x8B,
    thirteen=0x8C,
    fourteen=0x8D,
    fifteen=0x8E,
    sixteen=0x8F
)
MidiButtonType = Enum(Int8ul, toggle=0x80, trigger=0x81)
MixerDisplayBrightness = Enum(
    Int8ul,
    white=0x80,
    one=0x81,
    two=0x82,
    three=0x83,
    four=0x84,
    five=0x85)
MixerIndicatorBrightness = Enum(Int8ul, one=0x80, two=0x81, three=0x82)
ChannelFaderCurveLong = Enum(Int8ul, exponential=0x80, smooth=0x81, linear=0x82)


# 52 bytes
DjmMySettingBody = Struct(
    "u1" / Default(Bytes(12), b"xV4\x12\x01\x00\x00\x00 \x00\x00\x00"),
    "channel_fader_curve" / ChannelFaderCurve,
    "cross_fader_curve" / CrossfaderCurve,
    "headphones_pre_eq" / HeadphonesPreEQ,
    "headphones_mono_split" / HeadphonesMonoSplit,
    "beat_fx_quantize" / BeatFXQuantize,
    "mic_low_cut" / MicLowCut,
    "talk_over_mode" / TalkOverMode,
    "talk_over_level" / TalkOverLevel,
    "midi_channel" / MidiChannel,
    "midi_button_type" / MidiButtonType,
    "display_brightness" / MixerDisplayBrightness,
    "indicator_brightness" / MixerIndicatorBrightness,
    "channel_fader_curve_long" / ChannelFaderCurveLong,
    "u2" / Padding(27)  # Unknown values, always 0 (27 bytes)
)

# -- MySettings ------------------------------------------------------------------------

OnAirDisplay = Enum(Int8ul, off=0x80, on=0x81)
LCDBrightness = Enum(Int8ul, one=0x81, two=0x82, three=0x83, four=0x84, five=0x85)
Quantize = Enum(Int8ul, off=0x80, on=0x81)
AutoCueLevel = Enum(
    Int8ul,
    minus_36db=0x80,
    minus_42db=0x81,
    minus_48db=0x82,
    minus_54db=0x83,
    minus_60db=0x84,
    minus_66db=0x85,
    minus_72db=0x86,
    minus_78db=0x87,
    memory=0x88,
)
Language = Enum(
    Int8ul,
    english=0x81,
    french=0x82,
    german=0x83,
    italian=0x84,
    dutch=0x85,
    spanish=0x86,
    russian=0x87,
    korean=0x88,
    chinese_simplified=0x89,
    chinese_traditional=0x8A,
    japanese=0x8B,
    portuguese=0x8C,
    swedish=0x8D,
    czech=0x8E,
    hungarian=0x8F,
    danish=0x90,
    greek=0x91,
    turkish=0x92
)
JogRingBrightness = Enum(Int8ul, off=0x80, dark=0x81, bright=0x82)
JogRingIndicator = Enum(Int8ul, off=0x80, on=0x81)
SlipFlashing = Enum(Int8ul, off=0x80, on=0x81)
DiscSlotIllumination = Enum(Int8ul, off=0x80, dark=0x81, bright=0x82)
EjectLock = Enum(Int8ul, unlock=0x80, lock=0x81)
Sync = Enum(Int8ul, off=0x80, on=0x81)
PlayMode = Enum(Int8ul, continue_=0x80, single=0x81)
QuantizeBeatValue = Enum(Int8ul, one=0x80, half=0x81, quarter=0x82, eighth=0x83)
HotCueAutoLoad = Enum(Int8ul, off=0x80, on=0x81, rekordbox=0x82)
HotCueColor = Enum(Int8ul, off=0x80, on=0x81)
NeedleLock = Enum(Int8ul, unlock=0x80, lock=0x81)
TimeMode = Enum(Int8ul, elapsed=0x80, remain=0x81)
JogMode = Enum(Int8ul, cdj=0x80, vinyl=0x81)
AutoCue = Enum(Int8ul, off=0x80, on=0x81)
MasterTempo = Enum(Int8ul, off=0x80, on=0x81)
TempoRange = Enum(Int8ul, six=0x80, ten=0x81, sixteen=0x82, wide=0x83)
PhaseMeter = Enum(Int8ul, type1=0x80, type2=0x81)


# 40 bytes
MySettingBody = Struct(
    "u1" / Default(Bytes(8), b"xV4\x12\x02\x00\x00\x00"),
    "on_air_display" / OnAirDisplay,
    "lcd_brightness" / LCDBrightness,
    "quantize" / Quantize,
    "auto_cue_level" / AutoCueLevel,
    "language" / Language,
    "u2" / Default(Int8ul, 0x01),
    "jog_ring_brightness" / JogRingBrightness,
    "jog_ring_indicator" / JogRingIndicator,
    "slip_flashing" / SlipFlashing,
    "u3" / Default(Bytes(3), b"\x01\x01\x01"),
    "disc_slot_illumination" / DiscSlotIllumination,
    "eject_lock" / EjectLock,
    "sync" / Sync,
    "play_mode" / PlayMode,
    "quantize_beat_value" / QuantizeBeatValue,
    "hotcue_autoload" / HotCueAutoLoad,
    "hotcue_color" / HotCueColor,
    "u4" / Const(0, Int16ul),   # Unknown, always 0
    "needle_lock" / NeedleLock,
    "u5" / Const(0, Int16ul),   # Unknown, always 0
    "time_mode" / TimeMode,
    "jog_mode" / JogMode,
    "auto_cue" / AutoCue,
    "master_tempo" / MasterTempo,
    "tempo_range" / TempoRange,
    "phase_meter" / PhaseMeter,
    "u6" / Const(0, Int16ul),
)


# -- MySettings2 -----------------------------------------------------------------------

VinylSpeedAdjust = Enum(Int8ul, touch_release=0x80, touch=0x81, release=0x82)
JogDisplayMode = Enum(Int8ul, auto=0x80, info=0x81, simple=0x82, artwork=0x83)
PadButtonBrightness = Enum(Int8ul, one=0x81, two=0x82, three=0x83, four=0x84)
JogLCDBrightness = Enum(Int8ul, one=0x81, two=0x82, three=0x83, four=0x84, five=0x85)
WaveformDivisions = Enum(Int8ul, time_scale=0x80, phrase=0x81)
Waveform = Enum(Int8ul, waveform=0x80, phase_meter=0x81)
BeatJumpBeatValue = Enum(
    Int8ul,
    half=0x80,
    one=0x81,
    two=0x82,
    four=0x83,
    eight=0x84,
    sixteen=0x85,
    thirtytwo=0x86,
    sixtyfour=0x87
)


# 40 bytes
MySetting2Body = Struct(
    "vinyl_speed_adjust" / VinylSpeedAdjust,
    "jog_display_mode" / JogDisplayMode,
    "pad_button_brightness" / PadButtonBrightness,
    "jog_lcd_brightness" / JogLCDBrightness,
    "waveform_divisions" / WaveformDivisions,
    "u1" / Padding(5),
    "waveform" / Waveform,
    "u2" / Default(Int8ul, 0x81),
    "beat_jump_beat_value" / BeatJumpBeatValue,
    "u3" / Padding(27),   # Unknown (27 bytes), always 0
)


# -- DevSettings -----------------------------------------------------------------------

# 32 bytes
DevSettingBody = Struct(
    "u1" / Default(Bytes(8), b"xV4\x12\x02\x00\x00\x00"),
    "entries" / Bytes(24),
)


# -- Main Tags -------------------------------------------------------------------------


DjmMySetting = Struct(
    "len_strings" / Const(0x0060, Int32ul),
    "brand" / Const("PioneerDJ", PaddedString(32, "ascii")),
    "software" / Const("rekordbox", PaddedString(32, "ascii")),
    "version" / Const("1.000", PaddedString(32, "ascii")),
    "len_data" / Const(52, Int32ul),
    # Main data
    "data" / DjmMySettingBody,
    # Footer
    "checksum" / Int16ul,
    "unknown" / Const(0x00, Int16ul),
)


MySetting = Struct(
    "len_strings" / Const(0x0060, Int32ul),
    "brand" / Const("PIONEER", PaddedString(32, "ascii")),
    "software" / Const("rekordbox", PaddedString(32, "ascii")),
    "version" / Const("0.001", PaddedString(32, "ascii")),
    "len_data" / Const(40, Int32ul),
    # Main data
    "data" / MySettingBody,
    # Footer
    "checksum" / Int16ul,
    "unknown" / Const(0x00, Int16ul),
)


MySetting2 = Struct(
    "len_strings" / Const(0x0060, Int32ul),
    "brand" / Const("PIONEER", PaddedString(32, "ascii")),
    "software" / Const("rekordbox", PaddedString(32, "ascii")),
    "version" / Const("0.001", PaddedString(32, "ascii")),
    "len_data" / Const(40, Int32ul),
    # Main data
    "data" / MySetting2Body,
    # Footer
    "checksum" / Int16ul,
    "unknown" / Const(0x00, Int16ul),
)


DevSetting = Struct(
    "len_strings" / Const(0x0060, Int32ul),
    "brand" / Const("PIONEER DJ", PaddedString(32, "ascii")),
    "software" / Const("rekordbox", PaddedString(32, "ascii")),
    "version" / PaddedString(32, "ascii"),
    "len_data" / Const(32, Int32ul),
    # Main data
    "data" / DevSettingBody,
    # Footer
    "checksum" / Int16ul,
    "unknown" / Const(0x00, Int16ul),
)



# -*- coding: utf-8 -*-
# Author: Dylan Jones
# Date:   2022-10-24

"""Binary structures of Rekordbox ANLZ-files.

References
----------
.. [1] Rekordbox Export Structure Analysis: Analysis Files,
   https://djl-analysis.deepsymmetry.org/rekordbox-export-analysis/anlz.html
"""

from construct import Int8ub, Int16ub, Int32ub, PaddedString, Int32sb, StopIf
from construct import Const, Array, Padding, Bytes
from construct import Default, Enum, GreedyRange, Struct, Switch, this


# -- Beat Grid Tag (PQTZ) --------------------------------------------------------------

AnlzQuantizeTick = Struct(
    "beat" / Int16ub,
    "tempo" / Int16ub,
    "time" / Int32ub  # in ms from start
)

# len_header: 24
PQTZ = Struct(
    "u1" / Padding(4),
    "u2" / Const(0x80000, Int32ub),
    "entry_count" / Int32ub,
    "entries" / Array(this.entry_count, AnlzQuantizeTick),
)


# Extended Beat Grid Tag (PQT2)

AnlzQuantizeTick2 = Struct(
    "beat" / Int8ub,    # 1 byte
    "unkown" / Int8ub,    # 1 byte
)

# len_header: 56
PQT2 = Struct(
    Padding(4), # -> 16
    "u1" / Const(0x01000002, Int32ub),  # -> 20, count of "bpm" objects?
    Padding(4), # -> 24
    "bpm" / Array(2, AnlzQuantizeTick),  # -> 40 (2 * 8 bytes = 16 bytes)
    "entry_count" / Int32ub,  # -> 44 number of entries of 2 bytes
    "u3" / Int32ub,  # -> 48
    "u4" / Int32ub,  # -> 52
    "u5" / Int32ub,  # -> 56
    # End header
    StopIf(this.entry_count == 0),
    # "entries" / Array(this.entry_count, Bytes(2)),
    "entries" / Array(this.entry_count, AnlzQuantizeTick2),
)


# -- Cue List Tag (PCOB) ---------------------------------------------------------------

AnlzCuePointType = Enum(Int8ub, single=1, loop=2)

AnlzCuePointStatus = Enum(Int32ub, disabled=0, enabled=4)

AnlzTagCueObjectType = Enum(Int32ub, memory=0, hotcue=1)

AnlzCuePoint = Struct(
    "type" / Const("PCPT", PaddedString(4, encoding="ascii")),
    "len_header" / Int32ub,
    "len_entry" / Int32ub,
    "hot_cue" / Int32ub,  # 0 for memory
    "status" / AnlzCuePointStatus,
    "u1" / Const(0x10000, Int32ub),
    "order_first" / Int16ub,  # 0xffff for first cue, 0,1,3 for next
    "order_last" / Int16ub,  # 1,2,3 for first, second, third cue, 0xffff for last
    "type" / AnlzCuePointType,
    Padding(1),
    "u2" / Const(1000, Int16ub),
    "time" / Int32ub,
    "loop_time" / Default(Int32ub, -1),
    Padding(16),
)

# len_header: 24
PCOB = Struct(
    "cue_type" / AnlzTagCueObjectType,
    "unk" / Int16ub,
    "count" / Int16ub,
    "memory_count" / Int32sb,
    "entries" / Array(this.count, AnlzCuePoint),
)


# Extended (nxs2) Cue List Tag (PCO2)

AnlzCuePoint2 = Struct(
    "type" / Const("PCP2", PaddedString(4, encoding="ascii")),
    "len_header" / Int32ub,
    "len_entry" / Int32ub,
    "hot_cue" / Int32ub,  # 0 for memory
    "type" / Int8ub,  # spotted: 0x010003e8 0x020003e8
    Padding(3),
    "time" / Int32ub,
    "loop_time" / Default(Int32ub, -1),
    "color_id" / Int8ub,
    Padding(7),
    "loop_enumerator" / Int16ub,
    "loop_denominator" / Int16ub,
    "len_comment" / Int32ub,
    "comment" / PaddedString(this.len_comment, encoding="utf-16-be"),
    "color_code" / Int8ub,
    "color_red" / Int8ub,
    "color_green" / Int8ub,
    "color_blue" / Int8ub,
    Padding(this.len_entry - 48 - this.len_comment),
)

# len_header: 20
PCO2 = Struct(
    "type" / AnlzTagCueObjectType,
    "count" / Int16ub,
    "unknown" / Int16ub,
    "entries" / Array(this.count, AnlzCuePoint2),
)


# -- Path Tag (PPTH) -------------------------------------------------------------------

# len_header: 16
PPTH = Struct(
    "len_path" / Int32ub,  # is 0 for some tag types
    "path" / PaddedString(this.len_path - 2, encoding="utf-16-be"),
    Padding(2),
)


# -- VBR Tag (PVBR) --------------------------------------------------------------------

# len_header: 16
PVBR = Struct(
    "u1" / Int32ub,
    "idx" / Array(400, Int32ub),
    "u2" / Int32ub
)


# -- (Tiny) Waveform Preview Tag (PWAV / PWV2) -----------------------------------------

# len_header: 20
PWAV = Struct(
    "len_preview" / Int32ub,  # is 0 for some tag types
    "unknown" / Const(0x10000, Int32ub),
    "entries" / Array(this.len_preview, Int8ub),
)

# len_header: 20
PWV2 = Struct(
    "len_preview" / Int32ub,  # is 0 for some tag types
    "unknown" / Const(0x10000, Int32ub),
    "entries" / Array(this.len_preview, Int8ub),
)


# -- Waveform Detail Tag (PWV3) --------------------------------------------------------

# len_header: 24
PWV3 = Struct(
    "len_entry_bytes" / Const(1, Int32ub),
    "len_entries" / Int32ub,
    "u1" / Const(0x00960000, Int32ub),
    "entries" / Array(this.len_entries, Int8ub),
)


# -- Waveform Color Preview Tag (PWV4) -------------------------------------------------

# len_header: 24
PWV4 = Struct(
    "len_entry_bytes" / Const(0x00000006, Int32ub),
    "len_entries" / Int32ub,
    "unknown" / Int32ub,
    "entries" / Bytes(this.len_entry_bytes * this.len_entries),
)


# -- Waveform Color Detail Tag (PWV5) --------------------------------------------------

# len_header: 24
PWV5 = Struct(
    "len_entry_bytes" / Const(0x00000002, Int32ub),
    "len_entries" / Int32ub,
    "unknown" / Int32ub,
    "entries" / Array(this.len_entries, Int16ub),
)

# -- Song Structure Tag (PSSI) ---------------------------------------------------------

SongStructureEntry = Struct(
    "index" / Int16ub,
    "beat" / Int16ub,
    "kind" / Int16ub,
    "u1" / Int8ub,
    "k1" / Int8ub,
    "u2" / Int8ub,
    "k2" / Int8ub,
    "u3" / Int8ub,
    "b" / Int8ub,
    "beat_2" / Int16ub,
    "beat_3" / Int16ub,
    "beat_4" / Int16ub,
    "u4" / Int8ub,
    "k3" / Int8ub,
    "u5" / Int8ub,
    "fill" / Int8ub,
    "beat_fill" / Int16ub,
)

# len_header: 32
PSSI = Struct(
    "len_entry_bytes" /  Const(24, Int32ub),
    "len_entries" / Int16ub,
    "mood" / Int16ub,
    "u1" / Bytes(6),
    "end_beat" / Int16ub,
    "u2" / Bytes(2),
    "bank" / Int8ub,
    "u3" / Bytes(1),
    "entries" / Array(this.len_entries, SongStructureEntry),
)

# -- PWV6 ------------------------------------------------------------------------------

# len_header: 20
PWV6 = Struct(
    "len_entry_bytes" /  Const(0x00000003, Int32ub),
    "len_entries" / Int32ub,
    "entries" /  Bytes(this.len_entry_bytes * this.len_entries),
)

# -- PWV7 ------------------------------------------------------------------------------

# len_header: 24
PWV7 = Struct(
    "len_entry_bytes" /  Const(0x00000003, Int32ub),
    "len_entries" / Int32ub,
    "unknown" / Const(0x00960000, Int32ub),
    "entries" / Bytes(this.len_entry_bytes * this.len_entries),
)

# -- PWVC ------------------------------------------------------------------------------

# len_header: 14
PWVC = Struct(
    "unknown" /  Int16ub,
    "data" / Array(3, Int16ub)
)


# -- Main Items ------------------------------------------------------------------------

AnlzFileHeader = Struct(
    "type" / PaddedString(4, encoding="ascii"),
    "len_header" / Int32ub,
    "len_file" / Int32ub,
    "u1" / Int32ub,
    "u2" / Int32ub,
    "u3" / Int32ub,
    "u4" / Int32ub,
)


AnlzTag = Struct(
    "type" / PaddedString(4, encoding="ascii"),
    "len_header" / Int32ub,
    "len_tag" / Int32ub,
    "content" / Switch(
        this.type,
        {
            "PQTZ": PQTZ,
            "PQT2": PQT2,
            "PCOB": PCOB,  # seen in both DAT and EXT files
            "PCO2": PCO2,  # seen in EXT files
            "PPTH": PPTH,
            "PVBR": PVBR,
            "PSSI": PSSI,  # seen in EXT files
            "PWAV": PWAV,
            "PWV2": PWV2,
            "PWV3": PWV3,  # seen in EXT files
            "PWV4": PWV4,  # seen in EXT files
            "PWV5": PWV5,  # seen in EXT files
            "PWV6": PWV6,  # seen in 2EX files
            "PWV7": PWV7,  # seen in 2EX files
            "PWVC": PWVC,  # seen in 2EX files
        },
        default=Bytes(this.len_tag - 12),
    ),
)
